/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */
package org.brownell.xml;

import java.io.*;

import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.misc.*;



/**
 * This class provides a driver which may be invoked from the command line
 * to make a SAX parser use a specified handler to process a document.  This
 * facilitates some common types of command line tools, such as parsing an
 * XML document in order test it for well formedness or validity.
 *
 * <P>The following system properties are used to control the parser used,
 * and the handler it is given:  <UL>
 *
 *	<li><b>org.xml.sax.parser</b> ... the parser </li>
 *	<li><b>org.brownell.xml.DoParse.handler</b> ... the handler </li>
 *
 * </UL>
 *
 * <P> The SAX1 handler interfaces (DocumentHandler, DTDHandler, and
 * ErrorHandler) are given to the parser if implemented by the handler class.
 * Similarly, the SAX2 handler interfaces (Lexicalhandler, DeclHandler) are
 * used if the handler class implements these (and the parser is a SAX2
 * parser which supports such handlers).
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.0 (23 June 1999)
 */
final public class DoParse
{
    private static final String
	handlerPrefix = "http://xml.org/sax/handlers/";

    private DoParse () { /* no instances allowed */ }

    /**
     * Command line invoker for this class; pass a filename or URL
     * as the single argument.
     */
    public static void main (String argv [])
    throws IOException
    {
	InputSource	input;
	int		exitStatus = 1;

	if (argv.length != 1) {
	    System.err.println ("Usage: DoParse [filename|URL]");
	    System.exit (1);
	}

	try {
	    //
	    // Get input source for specified document (or try)
	    //
	    try {
		// first try treating it like a file.
		File	file = new File (argv [0]);
		String	temp;

		// On JDK 1.2 and later, File.toURL() is simpler!
		temp = file.getAbsolutePath ();
		if (File.separatorChar != '/')
		    temp = temp.replace (File.separatorChar, '/');
		if (!temp.startsWith ("/"))
		    temp = "/" + temp;
		temp = "file:" + temp;
		if (!temp.endsWith ("/") && file.isDirectory ())
		    temp = temp + "/";
		
		argv [0] = temp;

	    } catch (Exception e) {
		// if that fails, it's either a URL or an error
	    }
	    input = new InputSource (argv [0]);

	    //
	    // Get the parser and handler
	    //
	    Parser		parser;
	    Configurable	parser2 = null;

	    parser = ParserFactory.makeParser ();
	    if (parser instanceof Configurable)
		parser2 = (Configurable) parser;

	    //
	    // ... and handler
	    //
	    String		handlerClassName = null;
	    Object		handler = null;

	    try {
		Class		handlerClass;

		handlerClassName = System.getProperty (
			"org.brownell.xml.DoParse.handler", null);
		if (handlerClassName != null) {
		    handlerClass = Class.forName (handlerClassName);
		    handler = handlerClass.newInstance ();
		}

	    } catch (Exception e) {
		System.err.println ("Can't instantiate handler class.");
		System.err.println ("  Handler:  " + handlerClassName);
		System.err.println ("  Error:    " + e.getClass ().getName ());
		System.err.println ("  Message:  " + e.getMessage ());
		System.exit (1);
	    }

	    if (handler instanceof DocumentHandler)
		parser.setDocumentHandler ((DocumentHandler) handler);
	    if (handler instanceof DTDHandler)
		parser.setDTDHandler ((DTDHandler) handler);
	    if (handler instanceof ErrorHandler)
		parser.setErrorHandler ((ErrorHandler) handler);

	    try {
		if (parser2 != null && handler instanceof LexicalHandler)
		    parser2.setProperty (handlerPrefix + "LexicalHandler",
			(LexicalHandler) handler);
	    } catch (Exception e) {
		System.err.println ("Parser ignoring LexicallHandler");
	    }
	    try {
		if (handler instanceof DeclHandler)
		    parser2.setProperty (handlerPrefix + "DeclHandler",
			(DeclHandler) handler);
	    } catch (Exception e) {
		System.err.println ("Parser ignoring DeclHandler");
	    }

	    //
	    // XXX much better ErrorHandler setup is needed; we want
	    // to be able to use this as a commandline validator
	    // (printing and continuing from validity errors) and
	    // otherwise to control the diagnostics better.
	    //

	    //
	    // Parse, then clean up (let handler finalize etc)
	    //
	    parser.parse (input);
	    exitStatus = 0;
	    input = null;
	    parser = null;
	    handler = null;
	    System.gc ();

	} catch (SAXParseException err) {
	    System.err.println ("** Parsing error" 
		+ ", line " + err.getLineNumber ()
		+ ", uri " + err.getSystemId ());
	    System.err.println("   " + err.getMessage ());
	    
	    Exception	x = err;
	    if (err.getException () != null)
		x = err.getException ();

	    // XXX don't always want to do this (most fatal error)
	    x.printStackTrace ();

	} catch (SAXException e) {
	    Exception	x = e;
	    if (e.getException () != null)
		x = e.getException ();
	    x.printStackTrace ();

	} catch (Throwable t) {
	    t.printStackTrace ();
	}

	System.exit (exitStatus);
    }

/*
    static class MyErrorHandler extends HandlerBase
    {
	// treat validation errors as fatal
	public void error (SAXParseException e)
	throws SAXParseException
	{
	    throw e;
	}

	// dump warnings too
	public void warning (SAXParseException err)
	throws SAXParseException
	{
	    System.err.println ("** Warning" 
		+ ", line " + err.getLineNumber ()
		+ ", uri " + err.getSystemId ());
	    System.err.println("   " + err.getMessage ());
	}
    }
*/

}
