/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package org.brownell.xml;

import java.io.IOException;

import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.misc.*;

import org.w3c.dom.*;


/**
 * Builds a DOM document from the output of a SAX 2.0 (or SAX 1.0)
 * parser, using a defaulted or specified DOM implementation and parser.  
 * For example, a validating XML parser can be connected with an XML DOM;
 * or an HTML parser could be connected to an HTML (or non-HTML) DOM.
 *
 * <p> Note that some of the information exposed through DOM is not of
 * general interest, and does not truly relate to the core semantic model
 * of XML as consisting of elements, attributes, text, and processing
 * instructions.  By default, this builder only exposes those core node types,
 * and will not create any "extra" nodes, such as those for comments or
 * ignorable whitespace. This behavior may be changed by using the
 * <b><code>setSavingExtraNodes()</code></b> method.  The most useful of
 * such nodes are probably comments, which are used in some legacy
 * environments, such as HTML/XHTML, to wrap content such as inlined CSS
 * style directives or scripting code.  </p>
 *
 * <p> As a rule, if you ignore "extra" nodes and all of the incomplete DOM
 * Level 1 DTD functionality, and provide a SAX2 parser, the main portability
 * issue your code may have is that some SAX nonvalidating SAX parsers will
 * report ignorable whitespace characters as normal character data. </p>
 *
 * <hr width='70%'>
 *
 * <p> The DOM implementation class used is specified either as a parameter
 * to a construction method, or is derived from the value of the
 * <b><code>org.brownell.xml.DomBuilder.Document</code></b> system property.
 * That class must provide a default constructor, which creates an object
 * conforming to the DOM Level 1 core "Document" API.</p>
 *
 * <p> The SAX parser used is either provided as a constructor parameter,
 * or is gotten from the standard SAX ParserFactory.  In the latter case,
 * that parser should be an XML parser, for reasons of portability.  The
 * parser used should be a SAX 2.0 parser; see below for details of how
 * the SAX 1.0 API does not provide information needed to support DOM
 * correctly (it's more than just hiding data for "extra" nodes). </p>
 *
 * <p> By providing a parser directly, the caller can ensure that it has been
 * properly configured.  For example, it might be set up to validate.  The
 * caller may set up the ErrorHandler, Locale, and EntityResolver.  This
 * builder will assign the SAX 1.0 Document and DTD handlers of the parser,
 * and for a SAX 2.0 parser will also assign Lexical and Decl handlers if
 * those are supported by the parser.  Other handlers, except for the error
 * handler and the entity resolver, are reserved for future use by this
 * builder. </p>
 *
 * <hr width='70%'>
 *
 * <p> Because of missing functionality in the DOM Level 1 APIs, the
 * following DTD-related functionality can't be supported by software, such
 * as this builder, which does not use proprietary extensions to the
 * DOM APIs. <ul>
 *
 *	<li> Creation of DOM <b>DocumentType</b> nodes, or their associated
 *	<b>Notation</b> and <b>Entity</b> nodes. No APIs exist in DOM to
 *	create such nodes.</li>
 *
 *	<li> Creation of DOM <b>EntityReference</b> nodes.  The related
 *	<b>Entity</b> nodes can't be created, and in any case the children
 *	can't be set to be "read only".</b>
 *
 *	<li> The <b>Attr.isSpecified</b> flag can't be assigned through the
 *	DOM API. </li>
 *
 *	<li> Resetting values of element attributes which have defaults,
 *	when those attributes are deleted.  No API exists in DOM to tell
 *	the DOM which attributes have defaults, and what they are. </b>
 *
 *	</ul></p>
 *
 * <p> Some other functionality is not available through SAX, even using the
 * SAX2 parser APIs, and so can't be provided here.   Some other functionality
 * (some affecting correctness of DOM data models) is only available in those
 * SAX2 APIs, and so can't be provided when a SAX1 parser is in use.  <ul>
 *
 *	<li> There is no way to detect which attribute values were specified
 *	in the document, versus being defaulted through a DTD.  (As noted
 *	above, there is a parallel issue with the DOM API.) </li>
 *
 *	<li> Entity references which are found within attributes aren't
 *	reported as such.  This means that "Attr" nodes will not have such
 *	references as children.  (The current SAX2 draft has some issues in
 *	this area.) </li>
 *
 *	<li> <em>Unless using a SAX2 parser which reports them,</em>
 *	comments will not appear in the resulting document. </li>
 *
 *	<li> <em>Unless using a SAX2 parser which reports them,</em> CDATA
 *	delimiters for text will not cause CDATASection nodes to appear. </li>
 *
 *	<li><em>Unless using a SAX2 parser which reports DTD boundaries</em>,
 *	processing instructions (and perhaps comments, see above) found in
 *	a document's DTD will inappropriately be reported as being part of
 *	the document.</li>
 *
 *	</ul></p>
 *
 * <hr width='70%'>
 *
 * <p> Using proprietary DOM builder APIs may well be faster than using this
 * class and public APIs, because those proprietary APIs can eliminate some
 * of the conversion costs (e.g. character arrays to strings, and often back
 * again to character arrays) incurred by this class, as well as avoid
 * repeating certain integrity tests (such as ensuring that names are legal
 * XML 1.0 names). </p>
 *
 * <p> At this time, only XML 1.0 conformance is relied on; documents using
 * XML Namespaces are treated just like any other XML 1.0 document, and no
 * checks are made for conformance with the optional namespace specification.
 * Such additional checks would include:  rejecting entity, notation, and
 * processing instruction names with colons; rejecting ID, IDREF(S),
 * NOTATION, and ENTITY(IES) attribute values with colons; rejecting certain
 * legal XML 1.0 names, such as <em>:::</em> or <em>:foo:bar</em>; requiring
 * namespace prefixes to be declared; and ensuring that two values for one
 * attribute aren't provided (using different namespace prefixes). </p>
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.0 (30 June 1999)
 */
public class DomBuilder
{
    private Class		documentClass;
    private Parser		parser;
    private Handler		handler = new Handler ();

    private boolean		saveExtra;
    private Document		document;


    /**
     * Convenience routine, which uses the default DOM and parser (as
     * described above) to parse the specified (XML) document into a DOM
     * document tree.
     *
     * @param uri Identifies the resource to be parsed.
     *
     * @exception ClassNotFoundException The default DOM implementation
     *  or SAX parser class could not be found.  Check your class path
     *  and the system property values.
     * @exception NullPointerException The default DOM implementation or
     *  SAX parser name was null.  Check the system property values.
     * @exception IllegalAccessException The default DOM implementation
     *  or SAX parser class could not be loaded.
     * @exception InstantiationException The DOM implementation class or SAX
     *  parser class was successfully loaded, but could not be instantiated.
     *  Perhaps it has no default constructor.
     * @exception ClassCastException The default DOM implementation class
     *  was loaded and instantiated, but it does not implement the DOM
     *  Document interface.  <b>Or</b> the default SAX parser class was
     *  loaded and instantiated, but it does not implement the SAX 1.0 parser
     *  interface.
     */
    public static Document createDocument (String uri)
    throws SAXException, IOException, DOMException,
    	IllegalAccessException, InstantiationException,
	ClassNotFoundException
    {
	return createDocument (new InputSource (uri));
    }


    /**
     * Convenience routine, which uses the default DOM and parser (as
     * described above) to parse the specified (XML) document into a DOM
     * document tree.
     *
     * @param input Provided to the SAX parser as input.
     *
     * @exception ClassNotFoundException The default DOM implementation
     *  or SAX parser class could not be found.  Check your class path
     *  and the system property values.
     * @exception NullPointerException The default DOM implementation or
     *  SAX parser name was null.  Check the system property values.
     * @exception IllegalAccessException The default DOM implementation
     *  or SAX parser class could not be loaded.
     * @exception InstantiationException The DOM implementation class or SAX
     *  parser class was successfully loaded, but could not be instantiated.
     *  Perhaps it has no default constructor.
     * @exception ClassCastException The default DOM implementation class
     *  was loaded and instantiated, but it does not implement the DOM
     *  Document interface.  <b>Or</b> the default SAX parser class was
     *  loaded and instantiated, but it does not implement the SAX 1.0 parser
     *  interface.
     */
    public static Document createDocument (InputSource input)
    throws SAXException, IOException, DOMException,
    	IllegalAccessException, InstantiationException,
	ClassNotFoundException
    {
	DomBuilder	builder = new DomBuilder ();
	return builder.parse (input);
    }


    /**
     * Constructs a builder using the default DOM document class and the
     * default SAX parser.
     *
     * @exception ClassNotFoundException The default DOM implementation
     *  or SAX parser class could not be found.  Check your class path
     *  and the system property values.
     * @exception NullPointerException The default DOM implementation or
     *  SAX parser name was null.  Check the system property values.
     * @exception IllegalAccessException The default DOM implementation
     *  or SAX parser class could not be loaded.
     * @exception InstantiationException The DOM implementation class or SAX
     *  parser class was successfully loaded, but could not be instantiated.
     *  Perhaps it has no default constructor.
     * @exception ClassCastException The default DOM implementation class
     *  was loaded and instantiated, but it does not implement the DOM
     *  Document interface.  <b>Or</b> the default SAX parser class was
     *  loaded and instantiated, but it does not implement the SAX 1.0 parser
     *  interface.
     */
    public DomBuilder ()
    throws ClassNotFoundException, NullPointerException,
    	IllegalAccessException, InstantiationException, ClassCastException
    {
	this (
	    System.getProperty ("org.brownell.xml.DomBuilder.Document"),
	    ParserFactory.makeParser ());
    }


    /**
     * Constructs a builder using the specified DOM document class and the
     * default SAX parser.
     *
     * @param DOMDocumentClassName The name of the class implementing the
     *   kind of DOM document to be returned.
     *
     * @exception ClassNotFoundException The specified DOM implementation
     *  or default SAX parser class could not be found.  Check your class path
     *  and the values of those strings.
     * @exception NullPointerException The specified DOM implementation or
     *  default SAX parser name was null.  Check the values of those strings.
     * @exception IllegalAccessException The specified DOM implementation
     *  or default SAX parser class could not be loaded.
     * @exception InstantiationException The DOM implementation class or SAX
     *  parser class was successfully loaded, but could not be instantiated.
     *  Perhaps it has no default constructor.
     * @exception ClassCastException The specified DOM implementation class
     *  was loaded and instantiated, but it does not implement the DOM
     *  Document interface.  <b>Or</b> the default SAX parser class was
     *  loaded and instantiated, but it does not implement the SAX 1.0 parser
     *  interface.
     */
    public DomBuilder (String DOMDocumentClassName)
    throws ClassNotFoundException, NullPointerException,
    	IllegalAccessException, InstantiationException, ClassCastException
    {
	this (DOMDocumentClassName, ParserFactory.makeParser ());
    }


    /**
     * Constructs a builder using the default DOM document class and the
     * specified SAX parser.
     *
     * @param parser The SAX parser to be used; it may be partially
     *	configured, as described above.
     *
     * @exception ClassNotFoundException The default DOM implementation
     *  class could not be found.  Check your class path and the appropriate
     *  system property.
     * @exception NullPointerException The default DOM implementation class
     *  name was null.  Check the appropriate system property value.
     * @exception IllegalAccessException The default DOM implementation
     *  class could not be loaded.
     * @exception InstantiationException The DOM implementation class was
     *  successfully loaded, but could not be instantiated.  Perhaps it has
     *  no default constructor.
     * @exception ClassCastException The default DOM implementation class
     *  was loaded and instantiated, but it does not implement the DOM
     *  Document interface.
     */
    public DomBuilder (Parser parser)
    throws ClassNotFoundException, NullPointerException,
    	IllegalAccessException, InstantiationException, ClassCastException
    {
	this (
	    System.getProperty ("org.brownell.xml.DomBuilder.Document"),
	    parser);
    }


    /**
     * Constructs a builder using the specified DOM document class and the
     * specified SAX (2.0 or 1.0) parser.
     *
     * @param DOMDocumentClassName The name of the class implementing the
     *   kind of DOM document to be returned.
     * @param parser The SAX parser to be used; it may be partially
     *	configured, as described above.
     *
     * @exception ClassNotFoundException The specified DOM implementation
     *  class could not be found.  Check your class path.
     * @exception NullPointerException The specified DOM implementation class
     *  name was null.
     * @exception IllegalAccessException The specified DOM implementation
     *  class could not be loaded.
     * @exception InstantiationException The DOM implementation class was
     *  successfully loaded, but could not be instantiated.  Perhaps it has
     *  no default constructor.
     * @exception ClassCastException The specified DOM implementation class
     *  was loaded and instantiated, but it does not implement the DOM
     *  Document interface.
     */
    public DomBuilder (
	    String	DOMDocumentClassName,
	    Parser	parser
    ) throws ClassNotFoundException, NullPointerException,
    	IllegalAccessException, InstantiationException, ClassCastException
    {
	Object		temp;

	documentClass = Class.forName (DOMDocumentClassName);
	temp = documentClass.newInstance ();
	if (temp instanceof Document)
	    document = (Document) temp;
	else
	    throw new ClassCastException (
		"!((new " + DOMDocumentClassName
		+ " ()) instanceof org.w3c.dom.Document)");
	
	this.parser = parser;
	
	// set SAX1 handlers
	parser.setDocumentHandler (handler);
	parser.setDTDHandler (handler);

	// set SAX2 handlers as possible
	if (parser instanceof Configurable) {
	    Configurable	parser2 = (Configurable) parser;
	    try {
		parser2.setProperty (
		    "http://xml.org/sax/handlers/DeclHandler",
		    handler);
	    } catch (SAXException e) { /* IGNORE */ }
	    try {
		parser2.setProperty (
		    "http://xml.org/sax/handlers/LexicalHandler",
		    handler);
	    } catch (SAXException e) { /* IGNORE */ }
	}

	// user may provide (only) resolver, locale, and error handler
    }


    /**
     * Returns true if the builder is saving "extra" nodes, and false (the
     * default) otherwise.  "Extra" nodes are defined to be ignorable
     * whitespace, comments, and the use of CDATA nodes instead of normal
     * text nodes.  (Entity Reference nodes are also "extra", but can't be
     * exposed in any case since the DOM doesn't expose an API that permits
     * them to be constructed.)
     *
     * @see #setSavingExtraNodes
     */
    public boolean	isSavingExtraNodes ()
	{ return saveExtra; }


    /**
     * Controls whether the builder will save "extra" nodes.
     *
     * @see #isSavingExtraNodes
     * @param flag True iff extra nodes should be saved; false otherwise.
     */
    public void		setSavingExtraNodes (boolean flag)
	{ saveExtra = flag; }
    

    /**
     * Parses the specified document, returning its contents as a DOM
     * document.  This uses the SAX parser and DOM implementation which
     * were specified in the constructor to this builder.
     *
     * @param uri Identifies the resource to be parsed.
     * @exception SAXException As reported by the parser (in which case it is
     *   often a SAXParseException) or this builder.
     * @exception IOException As reported by the parser
     * @exception DOMException As reported by the DOM; always indicates
     *   a bug in either the DOM or the SAX parser
     */
    public Document	parse (String uri)
    throws SAXException, IOException, DOMException
    {
	return parse (new InputSource (uri));
    }

    
    /**
     * Parses the document provided, returning its contents as a DOM
     * document.  This uses the SAX parser and DOM implementation which
     * were specified in the constructor to this builder.
     *
     * @param input Provided to the SAX parser as input.
     * @exception SAXException As reported by the parser (in which case it is
     *   often a SAXParseException) or this builder.
     * @exception IOException As reported by the parser
     * @exception DOMException As reported by the DOM; always indicates
     *   a bug in either the DOM or the SAX parser
     */
    public Document	parse (InputSource input)
    throws SAXException, IOException, DOMException
    {
	// synchronizing, just in case caller violates mt-sanity rules
	synchronized (this) {
	    Document	retval;

	    if (document == null) {	// not 1st time through
		try {
		    document = (Document) documentClass.newInstance ();
		} catch (Exception e) {
		    // "can't happen" ... ideally not a SAX exception
		    throw new SAXException (
			"new " + documentClass.getName () + " ()", e);
		}
	    }
	    parser.parse (input);
	    retval = document;
	    document = null;
	    return retval;
	}
    }

    
    /**
     * Private inner class used to intercept various parsing events and use
     * them to populate a DOM document.
     *
     * <p> This is not offering ways to determine the base URI of the
     * resource from which the content was read ... so relative URIs in
     * external entity content can't ever be processed correctly. </p>
     *
     * <p> Note that if this ever throws a DOMException (runtime exception)
     * that will indicate a bug in the DOM (e.g. doesn't support something
     * per specification) or the parser (e.g. emitted an illegal name, or
     * accepted illegal input data). </p>
     */
    final class Handler extends DefaultHandler {
	private Node	top;
	private boolean	inCDATA;
	private boolean	mergeCDATA;
	private boolean	inDTD;

	// SAX1
	public void startDocument ()
	{
	    top = document;
	}

	// SAX1
	public void endDocument ()
	{
	    top = null;
	}

	// SAX1
	public void processingInstruction (String target, String data)
	{
	    ProcessingInstruction	pi;

	    if (inDTD)
		return;
	    pi = document.createProcessingInstruction (target, data);
	    top.appendChild (pi);
	}

	// SAX1
	public void characters (char buf [], int off, int len)
	{
	    String	value = new String (buf, off, len);
	    Node	lastChild = top.getLastChild ();

	    // merge consecutive text or CDATA nodes if appropriate.
	    if (lastChild instanceof Text) {
		if (!saveExtra
			// consecutive Text content ... always merge
			|| (!inCDATA
			    && !(lastChild instanceof CDATASection))
			// consecutive CDATASection content ... don't
			// merge between sections, only within them
			|| (inCDATA && mergeCDATA
			    && lastChild instanceof CDATASection)
			    ) {
		    ((Text)lastChild).appendData (value);
		    return;
		}
	    }
	    if (inCDATA && saveExtra) {
		top.appendChild (document.createCDATASection (value));
		mergeCDATA = true;
	    } else
		top.appendChild (document.createTextNode (value));
	}

	// SAX1
	public void startElement (String name, AttributeList attrs)
	{
	    Element	element = document.createElement (name);
	    int		length = attrs.getLength ();

	    for (int i = 0; i < length; i++)
		element.setAttribute (attrs.getName (i), attrs.getValue (i));
	    top.appendChild (element);
	    top = element;
	}

	// SAX1
	public void endElement (String name)
	{
	    top = top.getParentNode ();
	}

	// SAX1 (mandatory reporting if validating)
	public void ignorableWhitespace (char buf [], int off, int len)
	{
	    if (saveExtra)
		characters (buf, off, len);
	}

	// SAX2 lexical event
	public void startCDATA ()
	{
	    inCDATA = true;
	    // true except for the first fragment of a cdata section
	    mergeCDATA = false;
	}
	
	// SAX2 lexical event
	public void endCDATA ()
	{
	    inCDATA = false;
	}
	
	// SAX2 lexical event
	public void startDTD (String x, String y, String z)
	{
	    // need to filter out comments and PIs within the DTD
	    inDTD = true;
	}
	
	// SAX2 lexical event
	public void endDTD ()
	{
	    inDTD = false;
	}
	
	// SAX2 lexical event
	public void comment (char buf [], int off, int len)
	{
	    Comment	comment;

	    if (!saveExtra || inDTD)
		return;
	    comment = document.createComment (new String (buf, off, len));
	    top.appendChild (comment);
	}
    }
}
