/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package org.brownell.xml;

import java.io.IOException;
import java.util.Locale;

import org.xml.sax.*;
import org.xml.sax.helpers.AttributeListImpl;
import org.xml.sax.misc.DeclHandler;
import org.xml.sax.misc.LexicalHandler;

import org.w3c.dom.*;


/**
 * This parser emits SAX2 parsing events as it traverses a DOM tree, using
 * any conformant implementation of DOM.  It exposes all SAX 1.0 features,
 * and the following SAX 2 (1-June-1999 draft) features and properties (as
 * identified by standard URIs which are not fully provided here). </p>

 * <table border="1" width='100%' cellpadding='3' cellspacing='0'>
 * <tr bgcolor='#ccccff'>
 *	<th><font size='+1'>Name</font></th>
 *	<th><font size='+1'>Type</font></th>
 *	<th><font size='+1'>Notes</font></th></tr>
 * <tr><td>validation</td>
 *	<td>Feature</td>
 *	<td>false (does no parsing)</td></tr>
 * <tr><td>external-general-entities</td>
 *	<td>Feature</td>
 *	<td>false (does no parsing)</td></tr>
 * <tr><td>external-parameter-entities</td>
 *	<td>Feature</td>
 *	<td>false (does no parsing)</td></tr>
 * <tr><td>namespaces</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>false</em></td></tr>
 * <tr><td>normalize-text</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>false</em>; if you want normalized
 *	events for characters, you can get them by first normalizing
 *	the DOM tree.</td></tr>
 * <tr><td>use-locator</td>
 *	<td>Feature</td>
 *	<td>false (does no parsing)</td></tr>
 *
 * <tr><td colspan=3>&nbsp;</td></tr>
 *
 * <tr><td>dom-node</td>
 *	<td>Property</td>
 *	<td>This property may be set before parsing to hold a DOM
 *	<em>Document</em> node; any arguments given to <em>parse</em>
 *	methods are ignored.  When retrieved
 *	during a parse, this value contains the "current" DOM node.
 *	</td></tr>
 * <tr><td>DeclHandler</td>
 *	<td>Property</td>
 *	<td>A declaration handler may be provided.  Declaration of external
 *	general entities is exposed, but not parameter entities; none of the
 *	entity names reported here will begin with "%". </td></tr>
 * <tr><td>LexicalHandler</td>
 *	<td>Property</td>
 *	<td>A lexical handler may be provided.  While the start and end of
 *	any external subset are reported, expansion of other parameter
 *	entities (e.g. inside attribute list declarations) is not exposed.
 *	Expansion of general entities within attributes is also not exposed
 *	(see below).</td></tr>
 * </table>
 *
 * <P> Note that the current SAX2 draft has no standard way to provide the
 * initial DOM document.  This parser supports that facility through the
 * property holding the current DOM node:  it may be set when not parsing,
 * to a DOM Document node which is used to generate the events of the next
 * parse. </P>
 *
 * <P> The consequences of modifying a DOM document tree as it is being walked
 * by this "parser" are unspecified; don't do it! </P>
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.0 (23 June 1999)
 */
public class DomParser2 implements Configurable, Parser
{
    private static final String
	featurePrefix = "http://xml.org/sax/features/";

    // Stuff used internally to route events correctly
    private DefaultHandler	defaultHandler = new DefaultHandler ();

    // settable properties:
    private DeclHandler		declHandler = defaultHandler;
    private LexicalHandler	lexicalHandler = defaultHandler;
    private Document		document;

    // ... and SAX1 stuff
    private Locale		locale = Locale.getDefault ();
    private ErrorHandler	errHandler = defaultHandler;
    private DocumentHandler	docHandler = defaultHandler;
    private DTDHandler		dtdHandler = defaultHandler;
    private EntityResolver	resolver = defaultHandler;

    // current parser state:
    private Node		current;


    /**
     * Constructs an unitialized <b>SAX2</b> parser.
     */
    public DomParser2 () {
    } 

    /**
     * <b>SAX1</b>: Provides an object which receives callbacks for XML errors
     * of all levels (fatal, nonfatal, warning).
     */
    public void setErrorHandler (ErrorHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	errHandler = handler;
    }

    /**
     * <b>SAX1</b>: Provides an object which receives callbacks for the most
     * significant document information.
     */
    public void setDocumentHandler (DocumentHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	docHandler = handler;
    }

    /**
     * <b>SAX1</b>: Provides an object which may be used to intercept
     * declarations related to notations and unparsed entities.
     */
    public void setDTDHandler (DTDHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	dtdHandler = handler;
    }

    /**
     * <b>SAX1</b>: Provides an object which may be used when resolving external
     * entities during parsing (both general and parameter entities).
     */
    public void setEntityResolver (EntityResolver resolver)
    {
	if (resolver == null)
	    resolver = defaultHandler;
	this.resolver = resolver;
    }

    /**
     * <b>SAX1</b>: Identifies the locale which the parser should use for the
     * diagnostics it provides.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.setLocale()</em>
     */
    public void setLocale (Locale locale)
    throws SAXException
    {
	if (locale == null)
	    locale = Locale.getDefault ();
	this.locale = locale;
    }


    /**
     * <b>SAX1</b>:  Parses the previously provided DOM document (the
     * input parameter is ignored).
     *
     * @param uri ignored (pass an empty string)
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     * @exception IOException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     */
    public void parse (String uri) throws SAXException, IOException
    {
	walk ();
    }

    /**
     * <b>SAX1</b>:  Parses the previously provided DOM document (the
     * input parameter is ignored).
     *
     * @param input ignored
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     * @exception IOException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     */
    public void parse (InputSource input) throws SAXException, IOException
    {
	walk ();
    }

    /**
     * <b>SAX2</b>: Tells whether this parser supports the specified feature.
     * At this time, this directly parallels the underlying parser, except
     * that the use of validation can be enabled or disabled (since Sun's
     * package has two separate SAX 1.0 parsers).
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.getFeature()</em>
     */
    public boolean getFeature (String featureId)
    throws SAXException
    {
	// basically, none are supported -- they relate more to
	// parsing than to walking a "parse tree".
	if ((featurePrefix + "validation").equals (featureId)
		|| (featurePrefix + "external-general-entities")
		    .equals (featureId)
		|| (featurePrefix + "external-parameter-entities")
		    .equals (featureId)
		|| (featurePrefix + "namespaces").equals (featureId)
		|| (featurePrefix + "normalize-text").equals (featureId)
		|| (featurePrefix + "use-locator").equals (featureId)
		)
	    return false;

	throw new SAXNotRecognizedException (featureId);
    }

    /**
     * <b>SAX2</b>:  Returns the specified property.  At this time only
     * the declaration and lexical handlers, and current the "DOM" object,
     * are supported.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.getProperty()</em>
     */
    public Object getProperty (String propertyId)
    throws SAXException
    {
	if ("http://xml.org/sax/handlers/DeclHandler".equals (propertyId))
	    return declHandler;
	if ("http://xml.org/sax/handlers/LexicalHandler".equals (propertyId))
	    return lexicalHandler;
	if ("http://xml.org/sax/properties/dom-node".equals (propertyId))
	    return current;

	// unknown properties
	throw new SAXNotRecognizedException (propertyId);
    }

    /**
     * <b>SAX2</b>:  Sets the state of features supported in this parser.  At
     * this writing, no feature's state is mutable.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.setFeature()</em>
     */
    public void setFeature (String featureId, boolean state)
    throws SAXException
    {
	// see comments in getFeature ()

	if ((featurePrefix + "validation").equals (featureId)
		|| (featurePrefix + "external-general-entities")
		    .equals (featureId)
		|| (featurePrefix + "external-parameter-entities")
		    .equals (featureId)
		|| (featurePrefix + "namespaces").equals (featureId)
		|| (featurePrefix + "normalize-text").equals (featureId)
		|| (featurePrefix + "use-locator").equals (featureId)
		) {
	    if (state == false)
		return;
	    throw new SAXNotSupportedException (
		"feature is readonly:  " + featureId);
	}

	// unknown features
	throw new SAXNotRecognizedException (featureId);
    }

    /**
     * <b>SAX2</b>:  Assigns the specified property.  At this time only
     * declaration and lexical handlers, and the initial DOM document, are
     * supported.  These must not be changed to values of the wrong type.
     * Like SAX1 handlers, these handlers may be changed at any time.
     * Like SAX1 input source or document URI, the initial DOM document
     * may not be changed during a parse.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.setProperty()</em>
     */
    public void setProperty (String propertyId, Object property)
    throws SAXException
    {
	if ("http://xml.org/sax/handlers/DeclHandler".equals (propertyId)) {
	    if (!(property instanceof DeclHandler))
		throw new SAXNotSupportedException (propertyId);
	    declHandler = (DeclHandler) property;
	    return;
	}

	if ("http://xml.org/sax/handlers/LexicalHandler".equals (propertyId)) {
	    if (!(property instanceof LexicalHandler))
		throw new SAXNotSupportedException (propertyId);
	    lexicalHandler = (LexicalHandler) property;
	    return;
	}

	if ("http://xml.org/sax/properties/dom-node".equals (propertyId)) {
	    if (!(property instanceof Document))
		throw new SAXNotSupportedException (propertyId);
	    if (current != null)
		throw new SAXNotSupportedException (
		    "property is readonly during parse:  " + propertyId);
	    document = (Document) property;
	    return;
	}

	// unknown properties
	throw new SAXNotRecognizedException (propertyId);
    }

    //
    // Non-recursive walk, using DOM state when backtracking is needed
    //
    private void walk ()
    throws SAXException, IOException
    {
	int			type;
	NamedNodeMap		nodes;
	int			length;
	AttributeListImpl	attrs = new AttributeListImpl ();
	char			chars [];

	synchronized (this) {
	    if (current != null)
		throw new IllegalStateException ("already walking tree");

	    // JVM guarantees assignments are atomic; so no other
	    // thread could get this far till this walk's done.
	    current = document;
	}
    
	for (;;) {
	    type = current.getNodeType ();

	    //
	    // First, visit the current node, including any "start" calls
	    //
	    switch (type) {

	      case Node.DOCUMENT_NODE:
	        docHandler.startDocument ();
		break;

	      case Node.ELEMENT_NODE:
		nodes = current.getAttributes ();
		length = nodes.getLength ();
		for (int i = 0; i < length; i++) {
		    Attr attr = (Attr) nodes.item (i);

		    //
		    // NOTE:  DOM doesn't record the attribute type info
		    // which SAX exposes; this reports CDATA.
		    //
		    // NOTE:  SAX doesn't expose the isSpecified info which
		    // DOM exposes; that's discarded here.  Similarly with
		    // the information DOM hides inside itself about what
		    // the default values for an attribute are.
		    //
		    attrs.addAttribute (attr.getNodeName (),
			"CDATA", attr.getNodeValue ());
		}
	        docHandler.startElement  (current.getNodeName (), attrs);
		if (length != 0)
		    attrs.clear ();
		break;

	      case Node.CDATA_SECTION_NODE:
		lexicalHandler.startCDATA ();
		chars = current.getNodeValue ().toCharArray ();
		docHandler.characters (chars, 0, chars.length);
		lexicalHandler.endCDATA ();
		break;

	      case Node.COMMENT_NODE:
		chars = current.getNodeValue ().toCharArray ();
		lexicalHandler.comment (chars, 0, chars.length);
		break;

	      case Node.DOCUMENT_TYPE_NODE:
		{
		    DocumentType	doctype = (DocumentType) current;

		    //
		    // NOTE:  DOM doesn't support recreating either the
		    // external or internal subsets.
		    //
		    lexicalHandler.startDTD (doctype.getName (),
			    null, null);

		    // report notations first
		    nodes = doctype.getNotations ();
		    length = nodes.getLength ();
		    for (int i = 0; i < length; i++) {
			Notation notation = (Notation) nodes.item (i);
			    dtdHandler.notationDecl (
				notation.getNodeName (),
				notation.getPublicId (),
				notation.getSystemId ());
		    }

		    // then parsed and unparsed external general entities
		    nodes = doctype.getEntities ();
		    length = nodes.getLength ();
		    for (int i = 0; i < length; i++) {
			Entity	entity = (Entity) nodes.item (i);
			String	notation = entity.getNotationName ();

			if (notation != null)
			    dtdHandler.unparsedEntityDecl (
				entity.getNodeName (),
				entity.getPublicId (),
				entity.getSystemId (),
				notation);
			else if (entity.getSystemId () != null)
			    declHandler.externalEntityDecl (
				entity.getNodeName (),
				entity.getPublicId (),
				entity.getSystemId ());
			
			//
			// NOTE:  DOM doesn't clearly provide internal
			// entity support; but in case someone tries to
			// fudge such support, we defend ourselves above.
			//
			// NOTE:  DOM doesn't expose parameter entities
			// (thank you thank you thank you thank you)
			//
		    }

		    //
		    // NOTE:  DOM Level 1 (or current level 2 drafts)
		    // doesn't expose element or attribute declarations,
		    // as exposed by this SAX2 draft.  Presumably that'll
		    // come through XML schemas.
		    //
		    lexicalHandler.endDTD ();
		}
		break;

	      case Node.ENTITY_REFERENCE_NODE:
		lexicalHandler.startEntity (current.getNodeName ());
		break;

	      case Node.PROCESSING_INSTRUCTION_NODE:
	        docHandler.processingInstruction (
		    current.getNodeName (), current.getNodeValue ());
		break;

	      case Node.TEXT_NODE:
		chars = current.getNodeValue ().toCharArray ();
		docHandler.characters (chars, 0, chars.length);
		break;

	      default:
		// e.g. fragments, entities, attributes
		throw new SAXException ("Illegal DOM Node type in Document:  "
		    +  current.getNodeType ());
	    }

	    //
	    // Then, pick the next node to visit.  If the next node isn't
	    // a child, an "end" call may be needed before moving on.
	    // If there's no next node, we're done.
	    //
	    Node		next;

	    switch (type) {
	      case Node.DOCUMENT_NODE:
	      case Node.ELEMENT_NODE:
	      case Node.ENTITY_REFERENCE_NODE:
		//
		// For elements that can have children, visit those
		// children before any siblings (i.e. depth first)
		// and after visiting this node (i.e. preorder)
		//
		next = current.getFirstChild ();
		if (next != null) {
		    current = next;
		    break;
		}
		//
		// Else treat this like other childless nodes, but
		// handle this node's "end" immediately.
		//
		callEnd (current);

		// FALLTHROUGH

	      case Node.CDATA_SECTION_NODE:
	      case Node.COMMENT_NODE:
	      case Node.DOCUMENT_TYPE_NODE:
	      case Node.ENTITY_NODE:
	      case Node.PROCESSING_INSTRUCTION_NODE:
	      case Node.TEXT_NODE:
		//
		// Use next sibling, if there is one.
		// Else, climb up a level (calling "end")
		//	until we find an ancestral sibling
		//	or until we we climb off the top (FINISH)
		//
		for (;;) {
		    if ((next = current.getNextSibling ()) != null)
			break;
		    current = current.getParentNode ();
		    if (current == null)
			return;
		    callEnd (current);
		}
		current = next;
		break;

	      default:
		throw new SAXException (
		    "Illegal DOM Node type found:  " +  current.getNodeType ());
	    }
	}
    }

    private void callEnd (Node node) throws SAXException
    {
	switch (node.getNodeType ()) {
	  case Node.DOCUMENT_NODE:
	    docHandler.endDocument ();
	    return;

	  case Node.ELEMENT_NODE:
	    docHandler.endElement (node.getNodeName ());
	    return;

	  case Node.ENTITY_REFERENCE_NODE:
	    lexicalHandler.endEntity (node.getNodeName ());
	    return;

	  default:
	    // only those three container types may ever be found
	    // directly inside a Document.
	    throw new SAXException (
		"Illegal DOM container type found:  "
			+  current.getNodeType ());
	}
    }
}
