/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package org.brownell.xml;

import java.io.CharConversionException;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;

import org.xml.sax.*;
import org.xml.sax.misc.*;
import org.xml.sax.helpers.AttributeListImpl;


/**
 * This class is a SAX handler which echoes all its input as a well formed
 * XML or XHTML document.  If driven using SAX2 events, this output will
 * include a recreated document type declaration, and will optionally
 * include recreated entity references.
 *
 * <p> Few of the methods on this class are intended for applications to
 * use directly.  Those methods are for the JavaBeans properties used
 * to enable XHTML format output, or to disable recreation of entity
 * references.
 *
 * <P> Note that any relative URIs in the source document, as found in
 * entity and notation declarations, should have been fully resolved by
 * the parser providing events to this handler.  This means that the
 * output text should only have fully resolved URIs, which may not be
 * the desired behavior in cases where later binding is desired.
 *
 * <hr> <P> Users should note that the current SAX2 draft has problems with
 * how it reports the beginning and end of entities, where those entities
 * are used to construct other elements such as declarations in a DTD,
 * attribute values, or conditional sections.  In the face of such usage,
 * entity references reported through this handler are nonsensical. </P>
 *
 * <P> Also, declarations for attributes of type NOTATION are not fully
 * reported by this SAX2 draft; the particular notations which are legal
 * are not yet reported.  This class reports the permitted notations as
 * being the single <em>.unknown.</em> notation. </P>
 *
 * <P> When used with the <em>Parser2</em> adapter found in this package,
 * those entity reporting issues are dealt with by not reporting any of
 * the problematic entity expansions.  In short, when used with that
 * adapter, the following holds: </P>
 *
 * <UL>
 *
 *	<LI> General Entity references are expanded within attribute values.
 *	Elsewhere they are reported as <em>&amp;entity;</em> references.</LI>
 *
 *	<LI> No parameter entity references will be reported, so that any
 *	internal subset which uses such references will not be correctly
 *	recreated.  (Sun's parser reports the literal text of any internal
 *	subset, but this SAX2 draft does not.) </LI>
 *
 * </UL>
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.1 (3 September 1999)
 */
public class EchoHandler
    implements DocumentHandler, LexicalHandler, DTDHandler, DeclHandler
{
    // text prints/escapes differently depending on context
    //	CTX_ENTITY ... entity literal value
    //	CTX_ATTRIBUTE ... attribute literal value
    //	CTX_CONTENT ... content of an element
    //	CTX_UNPARSED ... CDATA, comment, PI, names, etc
    private static final int	CTX_ENTITY = 1;
    private static final int	CTX_ATTRIBUTE = 2;
    private static final int	CTX_CONTENT = 3;
    private static final int	CTX_UNPARSED = 4;

    // XXX names (element, attribute, PI, notation, etc) are not
    // currently written out with range checks (escapeChars).

    private static String	eol;

    static {
	try {
	    eol = System.getProperty ("line.separator", "\n");
	} catch (SecurityException e) {
	    eol = "\n";
	}
    }

    private Writer		out;
    private boolean		inCDATA;
    private int			entityNestLevel;
    private int			elementNestLevel;

    private short		dangerMask;
    private StringBuffer	stringBuf;
    private Locator		locator;

    private boolean		expandingEntities;
    private boolean		xhtml;
    private boolean		startedDoctype;
    private String		encoding;


    /**
     * Constructs a handler which writes all input to System.out in the
     * UTF-8 encoding, and closes System.out when endDocument is called.
     * (Yes it's annoying that this throws an exception -- but there's
     * really no way around it, since it's barely possible a JDK may
     * exist somewhere that doesn't know how to emit UTF-8.)
     */
    public EchoHandler () throws IOException
	{ this (new OutputStreamWriter (System.out, "UTF8")); }

    /**
     * Constructs a handler which writes all input to the writer, and then
     * closes the writer when the document ends.  If an XML declaration is
     * written onto the output, and this class can determine the name of
     * the character encoding for this writer, that encoding name will be
     * included in the XML declaration.
     *
     * <P> See the description of the constructor which takes an encoding
     * name for imporant information about selection of encodings.
     *
     * @param writer XML text is written to this writer.
     */
    public EchoHandler (Writer writer)
    {
	this (writer, null);
    }

    /**
     * Constructs a handler which writes all input to the writer, and then
     * closes the writer when the document ends.  If an XML declaration is
     * written onto the output, this class will use the specified encoding
     * name in that declaration.  If no encoding name is specified, no
     * encoding name will be declared unless this class can otherwise
     * determine the name of the character encoding for this writer.
     *
     * <P> At this time, only the UTF-8 ("UTF8") and UTF-16 ("Unicode")
     * output encodings are fully lossless with respect to XML data.  If you
     * use any other encoding you risk having your data be silently mangled
     * on output, as the standard Java character encoding subsystem silently
     * maps non-encodable characters to a question mark ("?") and will not
     * report such errors to applications.
     *
     * <p> For a few cases the risk can be reduced. If the writer is a
     * java.io.OutputStreamWriter, and uses either the ISO-8859-1 ("8859_1",
     * "ISO8859_1", etc) or US-ASCII ("ASCII") encodings, content which
     * can't be encoded in those encodings will be written safely.  Where
     * relevant, the XHTML entity names will be used; otherwise, numeric
     * character references will be emitted.
     *
     * <P> However, there remain a number of cases where substituting such
     * entity or character references is not an option.  Such references are
     * not usable within a DTD, comment, PI, or CDATA section.  Neither may
     * they be used when element, attribute, entity, or notation names have
     * the problematic characters.
     *
     * @param writer XML text is written to this writer.
     * @param encoding if non-null, and an XML declaration is written,
     *	this is the name that will be used for the character encoding.
     */
    public EchoHandler (Writer writer, String encoding)
    {
	if (writer == null)
	    throw new IllegalArgumentException ("null writer");

	out = writer;
	if (encoding == null && writer instanceof OutputStreamWriter)
	    encoding = ((OutputStreamWriter)writer).getEncoding ();

	if (encoding != null) {
	    encoding = encoding.toUpperCase ();

	    // Use official encoding names where we know them,
	    // never the less common ones.  When using common
	    // encodings where we can easily tell if characters
	    // are out of range, we'll escape out-of-range
	    // characters using character refs for safety.

	    // I _think_ these are all the main synonyms for these!
	    if ("UTF8".equals (encoding)) {
		encoding = "UTF-8";
	    } else if ("US-ASCII".equals (encoding)
		    || "ASCII".equals (encoding)) {
		dangerMask = (short) 0xff80;
		encoding = "US-ASCII";
	    } else if ("ISO-8859-1".equals (encoding)
		    || "8859_1".equals (encoding)
		    || "ISO8859_1".equals (encoding)) {
		dangerMask = (short) 0xff00;
		encoding = "ISO-8859-1";
	    } else if ("UNICODE".equals (encoding)
		    || "UNICODE-BIG".equals (encoding)
		    || "UNICODE-LITTLE".equals (encoding)) {
		encoding = "UTF-16";
	    }
	    this.encoding = encoding;

	    if (dangerMask != 0)
		stringBuf = new StringBuffer ();
	}
    }

    // JavaBeans properties

    /**
     * Controls whether the output should attempt to follow the "transitional"
     * XHTML rules so that it meets the "HTML Compatibility Guidelines"
     * appendix in the XHTML specification.  A "transitional" Document Type
     * Declaration (DTD) is placed near the beginning of the output document,
     * instead of whatever DTD would otherwise have been placed there, and
     * XHTML empty elements are printed specially.  When writing text in
     * US-ASCII or ISO-8859-1 encodings, the predefined XHTML internal
     * entity names are used (in preference to character references) when
     * writing content characters which can't be expressed in those encodings.
     *
     * <p> When this option is enabled, it is the caller's responsibility
     * to ensure that the input is otherwise valid as XHTML.  Things to
     * be careful of in all cases, as described in the appendix referenced
     * above, include:  <ul>
     *
     *	<li> Element and attribute names must be in lower case, both
     *		in the document and in any CSS style sheet.
     *	<li> All XML constructs must be valid as defined by the XHTML
     *		"transitional" DTD (including all familiar constructs,
     *		even deprecated ones).
     *	<li> The root element must be "html".
     *	<li> Elements that must be empty (such as <em>&lt;br&gt;</em>
     *		must have no content.
     *	<li> Use both <em>lang</em> and <em>xml:lang</em> attributes
     *		when specifying language.
     *	<li> Similarly, use both <em>id</em> and <em>name</em> attributes
     *		when defining elements that may be referred to through
     *		URI fragment identifiers ... and make sure that the
     *		value is a legal NMTOKEN, since not all such HTML 4.0
     *		identifiers are legal in XML.
     *	<li> Be careful with character encodings; make sure you provide
     *		a <em>&lt;meta http-equiv="Content-type"
     *		content="text/xml;charset=..." /&gt;</em> element in
     *		the HTML "head" element, naming the same encoding
     *		used to create this handler.  Also, if that encoding
     *		is anything other than US-ASCII, make sure that if
     *		the document is given a MIME content type, it has
     *		a <em>charset=...</em> attribute with that encoding.
     *	</ul>
     *
     * <p> Additionally, some of the oldest browsers have additional
     * quirks, to address with guidlines such as: <ul>
     *
     *	<li> Processing instructions may be rendered, so avoid them.
     *	<li> Embedded style sheets and scripts should not contain XML
     *		markup delimiters:  &amp;, &lt;, and ]]&gt; are trouble.
     *	<li> Attribute values should not have line breaks or multiple
     *		consecutive white space characters.
     *	<li> Use no more than one of the deprecated (transitional)
     *		<em>&lt;isindex&gt;</em> elements.
     *	<li> Some boolean attributes (such as <em>compact, checked,
     *		disabled, readonly, selected,</em> and more) confuse
     *		some browsers, since they only understand minimized
     *		versions which are illegal in XML.
     *	</ul>
     *
     * <p> Also, some characteristics of the resulting output may be
     * a function of whether the document is accessed through a path
     * giving it a MIME content type of <em>text/html</em> rather
     * than a content type indicating XML (<em>application/xml</em>
     * or <em>text/xml</em>).
     */
    public void setXhtml (boolean value)
    {
	if (locator != null)
	    throw new IllegalStateException ("started parsing");
	xhtml = value;
    }

    /**
     * Returns true if the output attempts to echo the input following
     * "transitional" XHTML rules and matching the "HTML Compatibility
     * Guidelines" so that an HTML version 3 browser can read the output
     * as HTML; returns false (the default) othewise.
     */
    public boolean isXhtml ()
    {
	return xhtml;
    }

    /**
     * Controls whether the output text contains references to
     * entities (the default), or instead contains the expanded
     * values of those entities.
     */
    public void setExpandingEntities (boolean value)
    {
	if (locator != null)
	    throw new IllegalStateException ("started parsing");
	expandingEntities = value;
    }

    /**
     * Returns true if the output will have no entity references;
     * returns false (the default) otherwise.
     */
    public boolean isExpandingEntities ()
    {
	return expandingEntities;
    }

    // SAX1 DocumentHandler
    /** <b>SAX1</b>:  provides parser status information */
    public void setDocumentLocator (Locator l)
    {
	locator = l;
    }
    
    /** <b>SAX1</b>:  indicates the beginning of a parse */
    public void startDocument ()
    throws SAXException
    {
	try {
	    if (out == null)
		throw new IllegalStateException (
		    "Can't reuse EchoHandler.");

	    // Unless the data is in US-ASCII, we always write the XML
	    // declaration when we know the encoding.  US-ASCII won't
	    // normally get mangled by web server confusion about the
	    // character encodings used.  Plus, it's an easy way to
	    // ensure we can write XHTML that's unlikely to confuse
	    // elderly HTML parsers.

	    if (dangerMask != (short) 0xff80 && encoding != null) {
		out.write ("<?xml version='1.0'");
		out.write (" encoding='" + encoding + "'");
		out.write ("?>");
		out.write (eol);
	    }

	    if (xhtml) {
		out.write ("<!DOCTYPE html PUBLIC");
		out.write (eol);
		out.write ("\t'-//W3C//DTD XHTML 1.0 Transitional//EN'");
		out.write (eol);
		out.write ("\t'http://www.w3.org/TR/xhtml1/DTD/transitional.dtd'>");
		out.write (eol);
		out.write (eol);
	    }
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX1</b>:  indicates the completion of a parse */
    public void endDocument ()
    throws SAXException
    {
	try {
	    out.write (eol);
	    out.write (eol);
	    out.close ();
	    out = null;
	    locator = null;
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    static final String		transitionalNamespace =
	    "http://www.w3.org/TR/xhtml1/transitional";
    
    // XHTML elements declared as EMPTY print differently
    private static boolean isEmptyElementTag (String tag)
    {
	switch (tag.charAt (0)) {
	  case 'a':	return "area".equals (tag);
	  case 'b':	return "base".equals (tag)
			    || "basefont".equals (tag)
			    || "br".equals (tag);
	  case 'c':	return "col".equals (tag);
	  case 'f':	return "frame".equals (tag);
	  case 'h':	return "hr".equals (tag);
	  case 'i':	return "img".equals (tag)
			    || "input".equals (tag)
			    || "isindex".equals (tag);
	  case 'l':	return "link".equals (tag);
	  case 'm':	return "meta".equals (tag);
	  case 'p':	return "param".equals (tag);
	}
	return false;
    }

    /** <b>SAX1</b>:  indicates the start of an element */
    public void startElement (String name, AttributeList atts)
    throws SAXException
    {
	try {
	    elementNestLevel++;
	    if (entityNestLevel != 0)
		return;
	    out.write ('<');
	    out.write (name);

		// XXX it may be desirable to sort these by name
		// particularly when used with a file that's got
		// a DTD, attribute defaulting may populate each
		// element with a confusing number of attributes.

	    if (atts != null && atts.getLength () != 0)
		for (int i= 0; i < atts.getLength (); i++) {
		    out.write (' ');
		    out.write (atts.getName (i));
		    out.write ('=');
		    writeQuotedValue (atts.getValue (i), CTX_ATTRIBUTE);
		}
	    if (xhtml && isEmptyElementTag (name))
		out.write (" /");
	    out.write ('>');
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX1</b>:  indicates the end of an element */
    public void endElement (String name)
    throws SAXException
    {
	try {
	    elementNestLevel--;
	    if (entityNestLevel != 0)
		return;
	    if (xhtml && isEmptyElementTag (name))
		return;
	    out.write ("</");
	    out.write (name);
	    out.write ('>');
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX1</b>:  reports content characters */
    public void characters (char buf [], int off, int len)
    throws SAXException
    {
	try {
	    if (entityNestLevel != 0)
		return;
	    if (inCDATA)
		escapeChars (buf, off, len, CTX_UNPARSED);
	    else
		escapeChars (buf, off, len, CTX_CONTENT);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX1</b>:  reports ignorable whitespace */
    public void ignorableWhitespace (char buf [], int off, int len)
    throws SAXException
    {
	try {
	    if (entityNestLevel != 0)
		return;
	    // don't forget to map NL to CRLF, CR, etc
	    escapeChars (buf, off, len, CTX_CONTENT);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX1</b>:  reports a PI */
    public void processingInstruction (String name, String value)
    throws SAXException
    {
	// ancient HTML browsers might render these ... their loss.
	// to prevent:  "if (xhtml) return;".

	try {
	    if (entityNestLevel != 0)
		return;
	    out.write ("<?");
	    out.write (name);
	    out.write (' ');
	    escapeChars (value.toCharArray (), -1, -1, CTX_UNPARSED);
	    out.write ("?>");
	    if (elementNestLevel == 0)
		out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    // SAX2 LexicalHandler

    /** <b>SAX2</b>:  called before parsing CDATA characters */
    public void startCDATA ()
    throws SAXException
    {
	try {
	    inCDATA = true;
	    if (entityNestLevel == 0)
		out.write ("<![CDATA[");
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called after parsing CDATA characters */
    public void endCDATA ()
    throws SAXException
    {
	try {
	    inCDATA = false;
	    if (entityNestLevel == 0)
		out.write ("]]>");
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called when the doctype is partially parsed */
    public void startDTD (String root, String pubid, String sysid)
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    startedDoctype = true;
	    out.write ("<!DOCTYPE ");
	    out.write (root);
	    out.write (' ');
	    if (pubid != null)
		out.write ("PUBLIC '" + pubid + "' '" + sysid + "' ");
	    else if (sysid != null)
		out.write ("SYSTEM '" + sysid + "' ");
	    out.write ('[');
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called after the doctype is parsed */
    public void endDTD ()
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    out.write ("]>");
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called before parsing an entity */
    public void startEntity (String name)
    throws SAXException
    {
	try {
	    boolean	writeEOL = true;

	    // Predefined XHTML entities (for characters) will get
	    // mapped back later.  No other entities should pass.
	    if (xhtml || expandingEntities)
		return;

	    entityNestLevel++;
	    if (name.equals ("[dtd]"))
		return;
	    if (entityNestLevel != 1)
		return;
	    if (!name.startsWith ("%")) {
		writeEOL = false;
		out.write ('&');
	    }
	    out.write (name);
	    out.write (';');
	    if (writeEOL)
		out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called after parsing an entity */
    public void endEntity (String name)
    throws SAXException
    {
	if (xhtml || expandingEntities)
	    return;
	entityNestLevel--;
    }

    /** <b>SAX2</b>:  called when comments are parsed */
    public void comment (char buf [], int off, int len)
    throws SAXException
    {
	try {
	    if (entityNestLevel != 0)
		return;
	    out.write ("<!--");
	    escapeChars (buf, off, len, CTX_UNPARSED);
	    out.write ("-->");
	    if (elementNestLevel == 0)
		out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    // SAX1 DTDHandler

    /** <b>SAX1</b>:  called on notation declarations */
    public void notationDecl (String name, String pubid, String sysid)
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    // At this time, only SAX2 callbacks start these.
	    if (!startedDoctype)  {
		// XXX write to temporary buffer, and make the start
		// of the root element write these declarations.
		// Normal form puts notations first, in alpha order;
		// followed by unparsed entities, in alpha order.
		return;
	    }

	    if (entityNestLevel != 0)
		return;
	    out.write ("<!NOTATION " + name + " ");
	    if (pubid != null)
		out.write ("PUBLIC \"" + pubid + '"');
	    else
		out.write ("SYSTEM ");
	    if (sysid != null)
		out.write ('"' + sysid + '"');
	    out.write (">");
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX1</b>:  called on unparsed entity declarations */
    public void unparsedEntityDecl (String name,
	String pubid, String sysid,
	String notation)
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    // At this time, only SAX2 callbacks start these.
	    if (!startedDoctype)  {
		// XXX write to temporary buffer, and make the start
		// of the root element write these declarations.
		// Normal form puts notations first, in alpha order;
		// followed by unparsed entities, in alpha order.
		return;
	    }

	    if (entityNestLevel != 0)
		return;
	    out.write ("<!ENTITY " + name + " ");
	    if (pubid != null)
		out.write ("PUBLIC \"" + pubid + '"');
	    else
		out.write ("SYSTEM ");
	    out.write ('"' + sysid + '"');
	    out.write (" NDATA " + notation + ">");
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    // SAX2 DeclHandler

    /** <b>SAX2</b>:  called on attribute declarations */
    public void attributeDecl (String element, String name,
	    String type, String defaultType, String defaultValue)
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    if (entityNestLevel != 0)
		return;
	    out.write ("<!ATTLIST " + element + ' ' + name + ' ');
	    out.write (type);
	    out.write (' ');
	    if (type.equals ("NOTATION"))	// XXX SAX2 API draft bug
		out.write ("( .unknown. ) ");
	    if (defaultType != null)
		out.write (defaultType + ' ');
	    if (defaultValue != null) 
		writeQuotedValue (defaultValue, CTX_ATTRIBUTE);
	    out.write ('>');
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called on element declarations */
    public void elementDecl (String name, String model)
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    if (entityNestLevel != 0)
		return;
	    out.write ("<!ELEMENT " + name + ' ' + model + '>');
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called on external entity declarations */
    public void externalEntityDecl (String name, String pubid, String sysid)
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    if (entityNestLevel != 0)
		return;
	    out.write ("<!ENTITY " + name + ' ');
	    if (pubid != null)
		out.write ("PUBLIC \"" + pubid + '"');
	    else
		out.write ("SYSTEM ");
	    out.write ('"' + sysid + "\">");
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    /** <b>SAX2</b>:  called on internal entity declarations */
    public void internalEntityDecl (String name, String value)
    throws SAXException
    {
	if (xhtml)
	    return;
	try {
	    if (entityNestLevel != 0)
		return;
	    out.write ("<!ENTITY " + name + ' ');
	    writeQuotedValue (value, CTX_ENTITY);
	    out.write ('>');
	    out.write (eol);
	} catch (IOException e) {
	    throw new SAXException ("can't write", e);
	}
    }

    private void writeQuotedValue (String value, int code)
    throws IOException
    {
	char	buf [] = value.toCharArray ();
	int	off = 0, len = buf.length;

	out.write ('"');
	escapeChars (buf, off, len, code);
	out.write ('"');
    }
    
    // From "HTMLlat1x.ent" ... names of entities for ISO-8859-1
    // (Latin/1) characters, all codes:  160-255 (0xA0-0xFF).
    // Codes 128-159 have no assigned values.
    private static final String HTMLlat1x [] = {
	// 160
	"nbsp", "iexcl", "cent", "pound", "curren",
	"yen", "brvbar", "sect", "uml", "copy",

	// 170
	"ordf", "laquo", "not", "shy", "reg",
	"macr", "deg", "plusmn", "sup2", "sup3",

	// 180
	"acute", "micro", "para", "middot", "cedil",
	"sup1", "ordm", "raquo", "frac14", "frac12",

	// 190
	"frac34", "iquest", "Agrave", "Aacute", "Acirc",
	"Atilde", "Auml", "Aring", "AElig", "Ccedil",

	// 200
	"Egrave", "Eacute", "Ecirc", "Euml", "Igrave",
	"Iacute", "Icirc", "Iuml", "ETH", "Ntilde",

	// 210
	"Ograve", "Oacute", "Ocirc", "Otilde", "Ouml",
	"times", "Oslash", "Ugrave", "Uacute", "Ucirc",

	// 220
	"Uuml", "Yacute", "THORN", "szlig", "agrave",
	"aacute", "acirc", "atilde", "auml", "aring",

	// 230
	"aelig", "ccedil", "egrave", "eacute", "ecirc",
	"euml", "igrave", "iacute", "icirc", "iuml",

	// 240
	"eth", "ntilde", "ograve", "oacute", "ocirc",
	"otilde", "ouml", "divide", "oslash", "ugrave",

	// 250
	"uacute", "ucirc", "uuml", "yacute", "thorn",
	"yuml"
    };

    // From "HTMLsymbolx.ent" ... some of the symbols that
    // we can conveniently handle.  Entities for the Greek.
    // alphabet (upper and lower cases) are compact.
    private static final String HTMLsymbolx_GR [] = {
	// 913
	"Alpha", "Beta", "Gamma", "Delta", "Epsilon",
	"Zeta", "Eta", "Theta", "Iota", "Kappa",

	// 923
	"Lambda", "Mu", "Nu", "Xi", "Omicron",
	"Pi", "Rho", null, "Sigma", "Tau",

	// 933
	"Upsilon", "Phi", "Chi", "Psi", "Omega"
    };

    private static final String HTMLsymbolx_gr [] = {
	// 945
	"alpha", "beta", "gamma", "delta", "epsilon",
	"zeta", "eta", "theta", "iota", "kappa",

	// 955
	"lambda", "mu", "nu", "xi", "omicron",
	"pi", "rho", "sigmaf", "sigma", "tau",

	// 965
	"upsilon", "phi", "chi", "psi", "omega"
    };


    // General routine to write text and substitute predefined
    // entities as needed.
    private void escapeChars (char buf [], int off, int len, int code)
    throws IOException
    {
	String	esc;
	int	first = 0;

	if (off < 0) {
	    off = 0;
	    len = buf.length;
	}
	for (int i = 0; i < len; i++) {
	    char c = buf [off + i];
	    switch (c) {
	      // Note that CTX_ATTRIBUTE isn't explicitly tested here;
	      // all syntax delimiters are escaped in CTX_ATTRIBUTE,
	      // otherwise it's similar to CTX_CONTENT

	      // ampersand flags entity references; entity replacement
	      // text has unexpanded references, other text doesn't.
	      case '&':
		if (code == CTX_ENTITY || code == CTX_UNPARSED)
		    continue;
		esc = "amp";
		break;

	      // attributes and text may NOT have literal '<', but
	      // entities may have markup constructs
	      case '<':
		if (code == CTX_ENTITY || code == CTX_UNPARSED)
		    continue;
		esc = "lt";
		break;

	      // as above re markup constructs; but otherwise
	      // this is just for consistency
	      case '>':
		if (code == CTX_ENTITY || code == CTX_UNPARSED)
		    continue;
	        esc = "gt";
		break;

	      // needed when printing quoted attribute/entity values
	      case '"':
		if (code == CTX_CONTENT || code == CTX_UNPARSED)
		    continue;
		esc = "quot";
		break;
	      case '\'':
		if (code == CTX_CONTENT || code == CTX_UNPARSED)
		    continue;
		esc = "apos";
		break;

	      // make line ends work per host OS convention
	      case '\n':
		esc = eol;
		break;

	      // no other characters NEED special treatment ... though many
	      // output encodings (like ASCII) would benefit from using
	      // numeric character refs for some common characters.
	      default:
		//
		// The only other characters that ALWAYS cause trouble
		// are those that:
		//
		//   (a) Are never legal in XML ... detected by a
		//	range check, and we'll assume that only
		//	legal surrogate pairs are given to us.
		//
		//   (b) Can't be encoded in this encoding, and which
		//	can't get reference substitution (e.g. inside
		//	CDATA sections, names, PI data, etc) ... we
		//	detect a few of these cases via dangerMask
		//	and the context code.
		//
		// There is NO WAY to write these so the recipient
		// gets exactly the data that was passed to us.
		//
		if ((c > 0xfffd)
			|| ((c < 0x0020) && !((c == 0x0009)
				|| (c == 0x000A) || (c == 0x000D)))
			|| (((c & dangerMask) != 0)
			    && (code == CTX_UNPARSED))) {
		    esc = "U+" + Integer.toHexString (c);

		    // If we can't pretty it up, panic
		    if (code == CTX_UNPARSED)
			throw new CharConversionException (
				"Can't encode: " + esc);

		    // else make it presentable 
		    // XXX println in a library, yeech
		    System.err.println ("WARNING: character mangling, "
			+ "char code = "+ esc);
		    esc = "lt;" + esc + "&gt";
		    break;
		}

		if ((c & dangerMask) == 0)
		    continue;
		esc = null;

		// Avoid numeric refs where symbolic ones exist, as
		// symbolic ones make more sense to humans reading!
		if (xhtml) {
		    // all the HTMLlat1x.ent entities
		    // (all the "ISO-8859-1" characters)
		    if (c >= 160 && c <= 255)
			esc = HTMLlat1x [c - 160];

		    // not quite half the HTMLsymbolx.ent entities
		    else if (c >= 913 && c <= 937)
			esc = HTMLsymbolx_GR [c - 913];
		    else if (c >= 945 && c <= 969)
			esc = HTMLsymbolx_gr [c - 945];

		    else switch (code) {
			// all of the HTMLspecialx.ent entities
			case  338: esc = "OElig";	break;
			case  339: esc = "oelig";	break;
			case  352: esc = "Scaron";	break;
			case  353: esc = "scaron";	break;
			case  376: esc = "Yuml";	break;
			case  710: esc = "circ";	break;
			case  732: esc = "tilde";	break;
			case 8194: esc = "ensp";	break;
			case 8195: esc = "emsp";	break;
			case 8201: esc = "thinsp";	break;
			case 8204: esc = "zwnj";	break;
			case 8205: esc = "zwj";		break;
			case 8206: esc = "lrm";		break;
			case 8207: esc = "rlm";		break;
			case 8211: esc = "ndash";	break;
			case 8212: esc = "mdash";	break;
			case 8216: esc = "lsquo";	break;
			case 8217: esc = "rsquo";	break;
			case 8218: esc = "sbquo";	break;
			case 8220: esc = "ldquo";	break;
			case 8221: esc = "rdquo";	break;
			case 8222: esc = "bdquo";	break;
			case 8224: esc = "dagger";	break;
			case 8225: esc = "Dagger";	break;
			case 8240: esc = "permil";	break;
			case 8249: esc = "lsaquo";	break;
			case 8250: esc = "rsaquo";	break;
			case 8364: esc = "euro";	break;

			// the other HTMLsymbox.ent entities
			case  402: esc = "fnof";	break;
			case  977: esc = "thetasym";	break;
			case  978: esc = "upsih";	break;
			case  982: esc = "piv";		break;
			case 8226: esc = "bull";	break;
			case 8230: esc = "hellip";	break;
			case 8242: esc = "prime";	break;
			case 8243: esc = "Prime";	break;
			case 8254: esc = "oline";	break;
			case 8260: esc = "frasl";	break;
			case 8472: esc = "weierp";	break;
			case 8465: esc = "image";	break;
			case 8476: esc = "real";	break;
			case 8482: esc = "trade";	break;
			case 8501: esc = "alefsym";	break;
			case 8592: esc = "larr";	break;
			case 8593: esc = "uarr";	break;
			case 8594: esc = "rarr";	break;
			case 8595: esc = "darr";	break;
			case 8596: esc = "harr";	break;
			case 8629: esc = "crarr";	break;
			case 8656: esc = "lArr";	break;
			case 8657: esc = "uArr";	break;
			case 8658: esc = "rArr";	break;
			case 8659: esc = "dArr";	break;
			case 8660: esc = "hArr";	break;
			case 8704: esc = "forall";	break;
			case 8706: esc = "part";	break;
			case 8707: esc = "exist";	break;
			case 8709: esc = "empty";	break;
			case 8711: esc = "nabla";	break;
			case 8712: esc = "isin";	break;
			case 8713: esc = "notin";	break;
			case 8715: esc = "ni";		break;
			case 8719: esc = "prod";	break;
			case 8721: esc = "sum";		break;
			case 8722: esc = "minus";	break;
			case 8727: esc = "lowast";	break;
			case 8730: esc = "radic";	break;
			case 8733: esc = "prop";	break;
			case 8734: esc = "infin";	break;
			case 8736: esc = "ang";		break;
			case 8743: esc = "and";		break;
			case 8744: esc = "or";		break;
			case 8745: esc = "cap";		break;
			case 8746: esc = "cup";		break;
			case 8747: esc = "int";		break;
			case 8756: esc = "there4";	break;
			case 8764: esc = "sim";		break;
			case 8773: esc = "cong";	break;
			case 8776: esc = "asymp";	break;
			case 8800: esc = "ne";		break;
			case 8801: esc = "equiv";	break;
			case 8804: esc = "le";		break;
			case 8805: esc = "ge";		break;
			case 8834: esc = "sub";		break;
			case 8835: esc = "sup";		break;
			case 8836: esc = "nsub";	break;
			case 8838: esc = "sube";	break;
			case 8839: esc = "supe";	break;
			case 8853: esc = "oplus";	break;
			case 8855: esc = "otimes";	break;
			case 8869: esc = "perp";	break;
			case 8901: esc = "sdot";	break;
			case 8968: esc = "lceil";	break;
			case 8969: esc = "rceil";	break;
			case 8970: esc = "lfloor";	break;
			case 8971: esc = "rfloor";	break;
			case 9001: esc = "lang";	break;
			case 9002: esc = "rang";	break;
			case 9674: esc = "loz";		break;
			case 9824: esc = "spades";	break;
			case 9827: esc = "clubs";	break;
			case 9829: esc = "hearts";	break;
			case 9830: esc = "diams";	break;
		    }
		}

		// else escape with numeric char refs
		if (esc == null) {
		    stringBuf.setLength (0);
		    stringBuf.append ("#");
		    stringBuf.append ((int) c);
		    esc = stringBuf.toString ();

		    // XXX that doesn't write surrogate pairs correctly;
		    // they should work as one ref per pair
		}
		break;
	    }
	    if (i != first)
		out.write (buf, off + first, i - first);
	    first = i + 1;
	    if (esc == eol)
		out.write (esc);
	    else {
		out.write ('&');
		out.write (esc);
		out.write (';');
	    }
	}
	if (first < len)
	    out.write (buf, off + first, len - first);
    }
}
