/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package org.brownell.xml;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.Locale;

import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.HTMLEditorKit.ParserCallback;
import javax.swing.text.html.HTML.Tag;
import javax.swing.text.html.parser.ParserDelegator;

import org.xml.sax.*;
import org.xml.sax.misc.LexicalHandler;
import org.xml.sax.helpers.AttributeListImpl;




/**
 * This is a wrapper around the <b>javax.swing.text.html.parser.*</b> HTML
 * parser, implementing the 1-June-1999 draft <b>SAX2</b> interfaces.
 * On valid HTML, and much invalid or malformed HTML, it produces a stream
 * of SAX parsing events corresponding to the parse of the corresponding
 * (well formed) XHTML document.  Element and attribute names are uniformly
 * presented in lower case.  (The Level 1 HTML DOM spec seems to be exotic
 * in not adopting that convention.) </p>
 *
 * <p> Only one type of lexical event is reported:  comments are visible.
 * This is generally used with HTML to access inlined CSS comments which are
 * protected against browsers old enough that they don't understand what the
 * "style" tag means.  Expansions of built-in entities (such as "&amp;nbsp;")
 * or character references are accordingly not visible.</p>
 *
 * <p> This parser does not support dynamic modification of the input stream
 * to the parser, needed to fully support <em>&lt;script&gt;</em> tags which
 * use the DOM to splice new page content into documents as they load. </p>
 *
 * <p> Current (Swing 1.1) HTML parsing issues of note include:</p>
 *
 * <ul>
 *
 *	<li> Not all HTML 4.0 features are recognized or supported. </li>
 *
 *	<li> Some XHTML constructs are handled poorly; for example, XML
 *	declarations at the beginning of a document cause multiple "html"
 *	elements to be reported, processing instructions are treated like
 *	regular text except that their leading '&lt;' is removed, and
 *	attributes with names such as <em>xml:lang</em> are treated as
 *	two separate attributes. </li>
 *
 *	<li> The content of a <em>&lt;noframes&gt;</em> element is treated as
 *	text, and elements within it aren't parsed.</li>
 *
 *	<li> HTML <em>&lt;link&gt;</em> elements are handled wrong; in some
 *	cases their attributes are coupled to a subsequent element.</li>
 *
 *	<li> HTML <em>&lt;style&gt;</em> elements are handled wrong; their
 *	contents aren't reported when wrapped in comments, and as with the
 *	<em>link</em> elements, their attributes may be shifted. </li>
 *
 *	<li> Some elements (often non-HTML names) may be given an "endtag"
 *	attribute (value true) that is not found in the input document.
 *	This is often coupled with the element being split into two, separated
 *	by the content which should be inside that element.</li>
 *
 *	<li> An issue with non-ASCII characters is that even for a JDK
 *	that handles an encoding, the standard IANA character set identifiers
 *	may not be recognized by the JDK.  Such bugs are slowly getting
 *	fixed; as a rule, use at least JDK 1.1.7 to ensure you have access
 *	to encoding names such as "ASCII" and "US-ASCII" (for the widely
 *	used seven bit standard character set).  </li>
 *
 *	<li> Error handling is a sore spot with HTML.  This parser doesn't do
 *	a very consistent job of categorizing errors; most are reported as
 *	warnings.  Diagnostics aren't very useful.  Swing's HTML parser does
 *	not accept as much malformed HTML as many widely used web browsers,
 *	and may not interpret such malformed HTML as expected.  </li>
 *
 *	</ul>
 *
 * <P> This driver adds ignorable newlines at various locations where they
 * won't be confused with HTML content.  These may of course be ignored.  If
 * they are not ignored, they make the output of this parser be more easily
 * printed, since otherwise HTML files of all sizes will appear without line
 * breaks of any kind, and viewing the output of this parser will cause
 * trouble for most text editors.</P>
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.2 (4 September 1999)
 */
public final class HtmlParser implements org.xml.sax.Parser, Configurable
{
    // Stuff used internally to route events correctly
    private Adapter		adapter = new Adapter ();
    private DefaultHandler	defaultHandler = new DefaultHandler ();

    // settable properties:
    private LexicalHandler	lexicalHandler = defaultHandler;

    // ... and SAX1 stuff
    private ErrorHandler	errHandler = defaultHandler;
    private DocumentHandler	docHandler = defaultHandler;
    private DTDHandler		dtdHandler = defaultHandler;
    private EntityResolver	resolver = defaultHandler;

    // debugging
    private static final boolean	debug = false;
    private static final boolean	debugErrs = false;


    /**
     * Constructs a new HTML parser.
     */
    public HtmlParser () {}


    /**
     * <b>SAX1</b>: Provides an object which receives callbacks for HTML errors
     * of all levels (fatal, nonfatal, warning).
     *
     * <p> Note that this parser does not provide a consistent categorization
     * of errors according to the categories defined in the SAX API.  Most
     * problems are reported at the "warning" level, and even those few
     * validity related errors reported at the "nonfatal" level may not be
     * viewed as issues in all HTML environments.  No errors are reported as
     * "fatal".
     *
     * <p> Throwing an exception from an error handler may not work well.
     */
    public void setErrorHandler (ErrorHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	errHandler = handler;
    }

    /**
     * <b>SAX1</b>: Provides an object which receives callbacks for the most
     * significant document information.
     */
    public void setDocumentHandler (DocumentHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	docHandler = handler;
    }

    /**
     * <b>SAX1</b>: Provides an object which may be used to intercept
     * declarations related to notations and unparsed entities.
     *
     * <p><em>Not used by this parser.</em>
     */
    public void setDTDHandler (DTDHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	dtdHandler = handler;
    }

    /**
     * <b>SAX1</b>: Provides an object which may be used when resolving external
     * entities during parsing (both general and parameter entities).
     *
     * <p><em>Not used by this parser.</em>
     */
    public void setEntityResolver (EntityResolver resolver)
    {
	if (resolver == null)
	    resolver = defaultHandler;
	this.resolver = resolver;
    }

    /**
     * <b>SAX1</b>: Identifies the locale which the parser should use for the
     * diagnostics it provides.
     *
     * <p><em>Not used by this parser.</em>
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.setLocale()</em>
     */
    public void setLocale (Locale locale)
    throws SAXException
    {
	// ignoring this entirely for now 
    }

    /**
     * <b>SAX1</b>:  parse the HTML text in the given input source.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     * @exception IOException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     */
    public void parse (InputSource input) throws SAXException, IOException
    {
	Reader	reader = input.getCharacterStream ();

	if (reader == null) {
	    InputStream	in = input.getByteStream ();
	    String	encoding;

	    if (in != null) {
		encoding = input.getEncoding ();
		if (encoding == null)
		    encoding = "8859_1";
		reader = new InputStreamReader (in, encoding);

	    } else if (input.getSystemId () != null) {
		URLConnection	conn;

		conn = new URL (input.getSystemId ()).openConnection ();
		conn.connect ();
		in = conn.getInputStream ();
		encoding = getEncoding (conn.getContentType (), "8859_1");
		reader = new InputStreamReader (in, encoding);

	    } else
		throw new IllegalArgumentException ("InputSource");
	}

	docHandler.startDocument ();
	try {
	    new ParserDelegator ().parse (reader, adapter, true);
	} catch (Wrapper x) {
	    throw x.x;
	}
	docHandler.endDocument ();
    }

    // currently a dumb parsing algorithm that works "mostly" and handles
    //	foo/bar
    //	foo/bar;charset=ABC
    //	foo/bar;charset=ABC;otherAttr=DEF
    //	foo/bar;charset=ABC (comment);otherAttr=DEF
    //	foo/bar;charset= "ABC" (comment);otherAttr=DEF
    static private String getEncoding (String contentType, String defValue)
    {
	int	temp;
	String	encoding;

	// Assumes 'charset' is only an attribute name, not part
	// of a value, comment, or other attribute name
	// ALSO assumes no escaped values like "\;" or "\)"
	if ((temp = contentType.indexOf ("charset")) != -1) {
	    // strip out everything up to '=' ...
	    temp = contentType.indexOf ('=', temp);
	    if (temp == -1)
		return defValue;
	    encoding = contentType.substring (temp + 1);
	    // ... and any subsequent attributes
	    if ((temp = encoding.indexOf (';')) != -1)
		encoding = encoding.substring (0, temp);
	    // ... and any comments after value
	    if ((temp = encoding.indexOf ('(')) != -1)
		encoding = encoding.substring (0, temp);
	    // ... then whitespace, and any (double) quotes
	    encoding = encoding.trim ();
	    if (encoding.charAt (0) == '"')
		encoding = encoding.substring (1, encoding.length () - 1);
	} else
	    encoding = defValue;
	return encoding;
    }


    /**
     * <b>SAX1</b>:  Parse the HTML text at the given input URI.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     * @exception IOException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     */
    public void parse (String uri) throws SAXException, IOException
    {
	parse (new InputSource (uri));
    }

    /**
     * <b>SAX2</b>: Tells whether this parser supports the specified feature.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.getFeature()</em>
     */
    public boolean getFeature (String featureId)
    throws SAXException
    {
	// none of that XML stuff!
	if ("http://xml.org/sax/features/validation".equals (featureId)
		|| "http://xml.org/sax/features/external-general-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/external-parameter-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/namespaces".equals (featureId)
		|| "http://xml.org/sax/features/normalize-text"
		    .equals (featureId)
		|| "http://xml.org/sax/features/use-locator"
		    .equals (featureId)
		)
	    return false;

	throw new SAXNotRecognizedException (featureId);
    }

    /**
     * <b>SAX2</b>:  Returns the specified property.  At this time only
     * lexical handlers are supported.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.getProperty()</em>
     */
    public Object getProperty (String propertyId)
    throws SAXException
    {
	if ("http://xml.org/sax/handlers/LexicalHandler".equals (propertyId))
	    return lexicalHandler;

	// unknown properties
	throw new SAXNotRecognizedException (propertyId);
    }

    /**
     * <b>SAX2</b>:  Sets the state of features supported in this parser.
     * As of this writing, no feature's state may be changed from its
     * default value.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.setFeature()</em>
     */
    public void setFeature (String featureId, boolean state)
    throws SAXException
    {
	// none of that XML stuff!
	if ("http://xml.org/sax/features/validation".equals (featureId)
		|| "http://xml.org/sax/features/external-general-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/external-parameter-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/namespaces".equals (featureId)
		|| "http://xml.org/sax/features/normalize-text"
		    .equals (featureId)
		|| "http://xml.org/sax/features/use-locator"
		    .equals (featureId)
		) {
	    if (state == false)
		return;
	    throw new SAXNotSupportedException (featureId);
	}

	// unknown features
	throw new SAXNotRecognizedException (featureId);
    }

    /**
     * <b>SAX2</b>:  Assigns the specified property.  At this time only
     * lexical handlers are supported, and these must not be changed to
     * values of the wrong type.  Like SAX1 handlers, these may be changed
     * at any time.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.setProperty()</em>
     */
    public void setProperty (String propertyId, Object property)
    throws SAXException
    {
	if ("http://xml.org/sax/handlers/LexicalHandler".equals (propertyId)) {
	    if (!(property instanceof LexicalHandler))
		throw new SAXNotSupportedException (propertyId);
	    lexicalHandler = (LexicalHandler) property;
	    return;
	}

	// unknown properties
	throw new SAXNotRecognizedException (propertyId);
    }


    static final private char newline [] = { '\n' };

    //
    // Adapter between Swing parser's events, and SAX ones.
    // 
    class Adapter extends ParserCallback
    {
	private AttributeListImpl	atts = new AttributeListImpl ();
	private boolean			recovering;
    
	// public void flush () { }

	public void handleText (char data [], int pos)
	{
	    if (debug)
		System.err.println ("handleText");
	    if (recovering)
		return;

	    try {
		docHandler.characters (data, 0, data.length);
	    } catch (SAXException e) {
		recovering = true;
		throw new Wrapper (e);
	    }
	}

	public void comment (char data [], int pos)
	{
	    if (debug)
		System.err.println ("comment");
	    if (recovering)
		return;

	    try {
// XXX "pos" signifies ... ?
		lexicalHandler.comment (data, 0, data.length);
	    } catch (SAXException e) {
		recovering = true;
		throw new Wrapper (e);
	    }
	}

	public void handleStartTag (
	    Tag				tag,
	    MutableAttributeSet 	attributes,
	    int				pos
	) {
	    String			name = tag.toString ();

	    if (debug)
		System.err.println ("handleStartTag " + name);
	    if (recovering)
		return;

	    try {
		convertAttributes (attributes);
		docHandler.startElement (name, atts);
		if (addSpaceInside (name))
		    docHandler.ignorableWhitespace (newline, 0, 1);
		atts.clear ();
	    } catch (SAXException e) {
		recovering = true;
		throw new Wrapper (e);
	    }
	}

	public void handleEndTag (Tag tag, int pos)
	{
	    String			name = tag.toString ();

	    if (debug)
		System.err.println ("handleEndTag " + name);
	    if (recovering)
		return;

	    try {
		docHandler.endElement (name);
		if (addSpaceAfter (name))
		    docHandler.ignorableWhitespace (newline, 0, 1);
	    } catch (SAXException e) {
		recovering = true;
		throw new Wrapper (e);
	    }
	}

	// Workaround a bug in the swing parser, where some elements
	// are treated as two simple elements separated by unrelated
	// content instead of one element containing that content
	private boolean isSwingBrokenTag (String name)
	{
	    return "div".equals (name)
		    || "span".equals (name)
		    || "abbr".equals (name);
	}

	public void handleSimpleTag (
	    Tag				tag,
	    MutableAttributeSet 	attributes,
	    int				pos
	) {
	    String			name = tag.toString ();

	    if (debug)
		System.err.println ("handleSimpleTag " + name);
	    if (recovering)
		return;

	    try {
		convertAttributes (attributes);
		if (isSwingBrokenTag (name)) {
		    if ("true".equals (atts.getValue ("endtag"))) {
			atts.clear ();
			docHandler.endElement (name);
			if (addSpaceAfter (name))
			    docHandler.ignorableWhitespace (newline, 0, 1);
		    } else {
			docHandler.startElement (name, atts);
			atts.clear ();
		    }
		} else {
		    docHandler.startElement (name, atts);
		    atts.clear ();
		    docHandler.endElement (name);
		    if (addSpaceAfter (name))
			docHandler.ignorableWhitespace (newline, 0, 1);
		}
	    } catch (SAXException e) {
		recovering = true;
		throw new Wrapper (e);
	    }
	}

	private boolean addSpaceInside (String tag)
	{
	    return "body".equals (tag)
		    || "head".equals (tag)
		    || "center".equals (tag)
		    ;
	}

	private boolean addSpaceAfter (String tag)
	{
	    return "p".equals (tag)
		    || "li".equals (tag)
		    || "table".equals (tag)
		    || "tr".equals (tag)
		    || "td".equals (tag)
		    || "h1".equals (tag)
		    || "h2".equals (tag)
		    || "h3".equals (tag)
		    || "h4".equals (tag)
		    || "h5".equals (tag)
		    || "h6".equals (tag)
		    || "dt".equals (tag)
		    || "dd".equals (tag)
		    || "center".equals (tag)
		    || "map".equals (tag)
		    || "area".equals (tag)
		    || "tbody".equals (tag)
		    || "title".equals (tag)
		    || "meta".equals (tag)
		    || "link".equals (tag)
		    || "style".equals (tag)
		    ;
	}

	//
	// This gets called lots with invalid HTML (or even just HTML that
	// the Swing parser was surprised by) ... can we provide better
	// diagnostics?  HotJava can, with more or less the same parser,
	// so it should be doable.
	//
	public void handleError (String diagnostic, int pos)
	{
	    if (debug || debugErrs)
		System.err.println ("handleError: " + diagnostic);

	    try {
		SAXParseException	ex;

		ex = new SAXParseException (diagnostic,
		    null, null,
		    -1, -1
		    );

		// Categorization of these errors is not trustworthy; but
		// we'll at least try to flag validity issues as such.
		if (diagnostic.startsWith ("invalid."))
		    errHandler.error (ex);
		else
		    errHandler.warning (ex);

	    } catch (SAXException e) {
		recovering = true;
		throw new Wrapper (e);
	    }
	}

	private void convertAttributes (MutableAttributeSet attributes)
	{
	    // invariant:  attributes are clear on entry!!

	    if (attributes.getAttributeCount () != 0) {
		for (Enumeration e = attributes.getAttributeNames ();
			e.hasMoreElements ();
			) {
		    Object	key = e.nextElement ();
		    String	name = key.toString ();
		    String	value;

		    value = attributes.getAttribute (key).toString ();
		    atts.addAttribute (name, "CDATA", value);
		}
	    }
	}
    }

    static class Wrapper extends RuntimeException {
	SAXException	x;

	Wrapper (SAXException x) { this.x = x; }
    }
}
