/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is free software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package org.brownell.xml;

import java.io.IOException;
import java.util.Locale;

import org.xml.sax.*;
import org.xml.sax.misc.DeclHandler;
import org.xml.sax.misc.LexicalHandler;
import org.xml.sax.helpers.ConfigurableParserAdapter;

// requires Oracle v2 parser -- see http://www.oracle.com
import oracle.xml.parser.v2.*;


/**
 * This is a wrapper around the <b>oracle.xml.parser.v2.SAXParser</b>
 * SAX 1.0 parser, implementing 1-June-1999 draft <b>SAX2</b> interfaces.
 * The standardized SAX2 facilities exposed by this parser include all
 * SAX 1.0 features, as well as a number of additional features and
 * properties identified by standardized URIs.  Those features (but not
 * the associated URIs) are summarized below.</p>
 *
 * <table border="1" width='100%' cellpadding='3' cellspacing='0'>
 * <tr bgcolor='#ccccff'>
 *	<th><font size='+1'>Name</font></th>
 *	<th><font size='+1'>Type</font></th>
 *	<th><font size='+1'>Notes</font></th></tr>
 * <tr><td>validation</td>
 *	<td>Feature</td>
 *	<td>Defaults to <em>false</em>; may be set to <em>true</em></td></tr>
 * <tr><td>external-general-entities</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>true</em></td></tr>
 * <tr><td>external-parameter-entities</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>true</em></td></tr>
 * <tr><td>namespaces</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>false</em></td></tr>
 * <tr><td>normalize-text</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>false</em></td></tr>
 * <tr><td>use-locator</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>true</em></td></tr>
 * </table>
 *
 * <p> None of the new SAX2 handlers is supported. </p>
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.0 (2 September 1999)
 */
public class OracleDriver extends ConfigurableParserAdapter
{
    private SAXParser		parser;

    /**
     * Constructs a parser which by default does not validate.
     */
    public OracleDriver ()
    {
	this (new SAXParser ());
    }

    private OracleDriver (SAXParser p)
    {
	super (p);
	parser = p;
	parser.setValidationMode (false);
    }

    /**
     * <b>SAX2</b>: Tells whether this parser supports the specified feature.
     * At this time, this directly parallels the underlying parser, except
     * that the use of validation can be enabled or disabled.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.getFeature()</em>
     */
    public boolean getFeature (String featureId)
    throws SAXException
    {
	// validate or not, by choice of underlying parser
	if ("http://xml.org/sax/features/validation".equals (featureId))
	    return parser.getValidationMode ();

	// external entities (both types) are currently always included
	if ("http://xml.org/sax/features/external-general-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/external-parameter-entities"
		    .equals (featureId))
	    return true;

	// element/attribute names are as written in document; no mangling
	if ("http://xml.org/sax/features/namespaces".equals (featureId))
	    return false;

	// text isn't collected ... could do it, it'd cost extra copies
	if ("http://xml.org/sax/features/normalize-text".equals (featureId))
	    return false;

	// always provides a locator
	if ("http://xml.org/sax/features/use-locator".equals (featureId))
	    return true;

	throw new SAXNotRecognizedException (featureId);
    }

    /**
     * <b>SAX2</b>:  Sets the state of features supported in this parser.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.setFeature()</em>
     */
    public void setFeature (String featureId, boolean state)
    throws SAXException
    {
	// see comments in getFeature ()

	// features with mutable state:
	if ("http://xml.org/sax/features/validation".equals (featureId)) {
	    parser.setValidationMode (state);
	    return;
	}

	// features with immutable state:
	if ("http://xml.org/sax/features/external-general-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/external-parameter-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/use-locator"
		    .equals (featureId)) {
	    if (state != true)
		throw new SAXNotSupportedException (featureId);
	    else
		return;
	}
	if ("http://xml.org/sax/features/namespaces"
		    .equals (featureId)
		|| "http://xml.org/sax/features/normalize-text"
		    .equals (featureId)) {
	    if (state != false)
		throw new SAXNotSupportedException (featureId);
	    else
		return;
	}

	// unknown features
	throw new SAXNotRecognizedException (featureId);
    }
}
