/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is free software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package org.brownell.xml;

import java.io.IOException;
import java.util.Locale;

import org.xml.sax.*;
import org.xml.sax.misc.DeclHandler;
import org.xml.sax.misc.LexicalHandler;

// requires TR2 -- see http://java.sun.com/products/xml
import com.sun.xml.parser.DtdEventListener;
import com.sun.xml.parser.LexicalEventListener;
import com.sun.xml.parser.ValidatingParser;
// there's also a nonvalidating parser ("Parser") there (name collision)


/**
 * This is a wrapper around the <b>com.sun.xml.parser.*</b> SAX 1.0 parsers,
 * implementing 1-June-1999 draft <b>SAX2</b> interfaces.  The standardized
 * SAX2 facilities exposed by this parser include all SAX 1.0 features, as
 * well as a number of additional features and properties identified by
 * standardized URIs.  Those features (but not the associated URIs) are
 * summarized below.</p>
 *
 * <table border="1" width='100%' cellpadding='3' cellspacing='0'>
 * <tr bgcolor='#ccccff'>
 *	<th><font size='+1'>Name</font></th>
 *	<th><font size='+1'>Type</font></th>
 *	<th><font size='+1'>Notes</font></th></tr>
 * <tr><td>validation</td>
 *	<td>Feature</td>
 *	<td>Defaults to <em>false</em>; may be set to <em>true</em></td></tr>
 * <tr><td>external-general-entities</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>true</em></td></tr>
 * <tr><td>external-parameter-entities</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>true</em></td></tr>
 * <tr><td>namespaces</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>false</em></td></tr>
 * <tr><td>normalize-text</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>false</em></td></tr>
 * <tr><td>use-locator</td>
 *	<td>Feature</td>
 *	<td>Value is fixed at <em>true</em></td></tr>
 *
 * <tr><td colspan=3>&nbsp;</td></tr>
 *
 * <tr><td>DeclHandler</td>
 *	<td>Property</td>
 *	<td>A declaration handler may be provided.  Declaration of general
 *	entities is exposed, but not parameter entities; none of the entity
 *	names reported here will begin with "%". </td></tr>
 * <tr><td>LexicalHandler</td>
 *	<td>Property</td>
 *	<td>A lexical handler may be provided.  While the start and end of
 *	any external subset are reported, expansion of other parameter
 *	entities (e.g. inside attribute list declarations) is not exposed.
 *	Expansion of general entities within attributes is also not exposed
 *	(see below).</td></tr>
 * </table>
 *
 * <p> Other features of the Sun parser, which may be significant by their
 * current absence in the draft SAX2 APIs, include: <ul>
 *
 *	<li> The <em>Internal DTD Subset</em> is not currently exposed
 *	through this interface; applications requiring literal recreation of
 *	a document's doctype can't do this with SAX2.  (Examples of such
 *	applications include the filtering proxies that many secure message
 *	distribution networks require.)</li>
 *
 *	<li> Sun's parser lets you insert the results of executing the
 *	JavaScript within an HTML <em>&lt;script&gt;</em> tag into the input
 *	stream as it is parsed, permitting complete XHTML support. </li>
 *
 *	<li> SAX2 discards some semantic attribute declaration information,
 *	namely the list of just which NOTATIONs an attribute may use. </li>
 *
 *	<li> At this time, the SAX 1.0 handlers are not exposed as parser
 *	properties in the way that the SAX2 ones are exposed.  (These could
 *	be defined as such properties, however.)  </li>
 *
 *	<li> The <em>com.sun.xml.parser.AttributeListEx</em> information is
 *	still visible through the startElement() callback; this information
 *	is useful to support some DOM Level 1 features relating to defaulting
 *	of attributes, without requiring applications to replicate the
 *	parser's handling of DTD information.  </li>
 *
 * </ul>
 *
 * <p> Features which might affect conformance with the SAX2 draft include:
 * <ul>
 *
 *	<li> While validation may be enabled or disabled, that can't be done
 *	while parsing.  The SAX2 draft should perhaps say this, since any
 *	nonvalidating parser is likely to have discarded information needed
 *	to validate by the time validation would be enabled. </li>
 *
 *	<li> Like other SAX handlers, the declaration and lexical handlers may
 *	be set at any time ... though the SAX2 draft says that shouldn't be
 *	permitted. </li>
 *
 *	<li> Expansion of entities within attributes is not exposed.  It was
 *	observed during initial implementation of Sun's parser that such
 *	reports look like the entities precede the element, giving
 *	applications an incorrect model of the document content. </li>
 *
 *	<li> As noted above, Sun's parsers will not report expansion of
 *	parameter entities.  The rationale is, in short, that seeing such
 *	boundaries is all but useless since they're normally used as parts
 *	of larger components, and construction-by-parts isn't exposed. </li>
 *
 * </ul>
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.1 (23 June 1999)
 */
public class Parser2 implements Configurable, Parser
{
    // Stuff used internally to route events correctly
    private Adapter		adapter = new Adapter ();
    private DefaultHandler	defaultHandler = new DefaultHandler ();

    // either a validating or nonvalidating parser
    private Parser		parser;

    // settable features:
    private boolean		validation;

    // settable properties:
    private DeclHandler		declHandler = defaultHandler;
    private LexicalHandler	lexicalHandler = defaultHandler;

    // ... and SAX1 stuff
    private Locale		locale = Locale.getDefault ();
    private ErrorHandler	errHandler = defaultHandler;
    private DocumentHandler	docHandler = defaultHandler;
    private DTDHandler		dtdHandler = defaultHandler;
    private EntityResolver	resolver = defaultHandler;


    /**
     * Constructs an unitialized <b>SAX2</b> parser.  By default, this will
     * not validate; validation may be manually enabled.  If it is enabled,
     * you should probably set an error handler accordingly.
     */
    public Parser2 () {
    } 

    /**
     * <b>SAX1</b>: Provides an object which receives callbacks for XML errors
     * of all levels (fatal, nonfatal, warning).
     *
     * <p> <b>NOTE:</b>  When using a validating parser, the SAX default error
     * handler almost certainly does not do what you expected.  Unless you
     * specified validation on a lark, you probably wanted to do something
     * more with validity errors than discard them.  You should provide a
     * handler which does something useful with nonfatal errors, probably
     * reporting them (perhaps by treating them like fatal errors).
     */
    public void setErrorHandler (ErrorHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	errHandler = handler;
	if (parser != null)
	    parser.setErrorHandler (handler);
    }

    /**
     * <b>SAX1</b>: Provides an object which receives callbacks for the most
     * significant document information.
     */
    public void setDocumentHandler (DocumentHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	docHandler = handler;
	if (parser != null)
	    parser.setDocumentHandler (handler);
    }

    /**
     * <b>SAX1</b>: Provides an object which may be used to intercept
     * declarations related to notations and unparsed entities.
     */
    public void setDTDHandler (DTDHandler handler)
    {
	if (handler == null)
	    handler = defaultHandler;
	dtdHandler = handler;
	if (parser != null)
	    parser.setDTDHandler (handler);
    }

    /**
     * <b>SAX1</b>: Provides an object which may be used when resolving external
     * entities during parsing (both general and parameter entities).
     */
    public void setEntityResolver (EntityResolver resolver)
    {
	if (resolver == null)
	    resolver = defaultHandler;
	this.resolver = resolver;
	if (parser != null)
	    parser.setEntityResolver (resolver);
    }

    /**
     * <b>SAX1</b>: Identifies the locale which the parser should use for the
     * diagnostics it provides.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.setLocale()</em>
     */
    public void setLocale (Locale locale)
    throws SAXException
    {
	if (locale == null)
	    locale = Locale.getDefault ();
	this.locale = locale;
	if (parser != null)
	    parser.setLocale (locale);
    }

    private void initParser ()
    throws SAXException
    {
	if (validation)
	    parser = new ValidatingParser ();
	else
	    parser = new com.sun.xml.parser.Parser ();

	parser.setLocale (locale);
	parser.setErrorHandler (errHandler);
	parser.setEntityResolver (resolver);

	// convert (if needed) between Sun and SAX2 events
	parser.setDTDHandler (adapter);
	parser.setDocumentHandler (adapter);
    }

    /**
     * <b>SAX1</b>:  Parse the XML text at the given input URI.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     * @exception IOException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     */
    public void parse (String uri) throws SAXException, IOException
    {
	initParser ();
	parser.parse (uri);
    }

    /**
     * <b>SAX1</b>:  parse the XML text in the given input source.  Note that
     * if no system identifier is available for the input stream, then the
     * parse will not be able to interpret any relative URIs in that XML
     * source text; provide the system identifier, if you know it.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     * @exception IOException as defined in the specification for
     *	<em>org.xml.sax.Parser.parse()</em>
     */
    public void parse (InputSource input) throws SAXException, IOException
    {
	initParser ();
	parser.parse (input);
    }

    /**
     * <b>SAX2</b>: Tells whether this parser supports the specified feature.
     * At this time, this directly parallels the underlying parser, except
     * that the use of validation can be enabled or disabled (since Sun's
     * package has two separate SAX 1.0 parsers).
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.getFeature()</em>
     */
    public boolean getFeature (String featureId)
    throws SAXException
    {
	// validate or not, by choice of underlying parser
	if ("http://xml.org/sax/features/validation".equals (featureId))
	    return validation;

	// external entities (both types) are currently always included
	if ("http://xml.org/sax/features/external-general-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/external-parameter-entities"
		    .equals (featureId))
	    return true;

	// element/attribute names are as written in document; no mangling
	if ("http://xml.org/sax/features/namespaces".equals (featureId))
	    return false;

	// text isn't collected ... could do it, it'd cost extra copies
	if ("http://xml.org/sax/features/normalize-text".equals (featureId))
	    return false;

	// always provides a locator
	if ("http://xml.org/sax/features/use-locator".equals (featureId))
	    return true;

	throw new SAXNotRecognizedException (featureId);
    }

    /**
     * <b>SAX2</b>:  Returns the specified property.  At this time only
     * declaration and lexical handlers are supported (Sun's parser exposes
     * this state already).
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.getProperty()</em>
     */
    public Object getProperty (String propertyId)
    throws SAXException
    {
	if ("http://xml.org/sax/handlers/DeclHandler".equals (propertyId))
	    return declHandler;
	if ("http://xml.org/sax/handlers/LexicalHandler".equals (propertyId))
	    return lexicalHandler;

	// unknown properties
	throw new SAXNotRecognizedException (propertyId);
    }

    /**
     * <b>SAX2</b>:  Sets the state of features supported in this parser.  At
     * this writing, only one feature's state is mutable:  validation can be
     * enabled or disabled, though not in the middle of a parse.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.setFeature()</em>
     */
    public void setFeature (String featureId, boolean state)
    throws SAXException
    {
	// see comments in getFeature ()

	// features with mutable state:
	if ("http://xml.org/sax/features/validation".equals (featureId)) {
	    if (parser != null) 
		throw new SAXNotSupportedException (
		    "feature is readonly during parse:  " + featureId);
	    validation = state;
	    return;
	}

	// features with immutable state:
	if ("http://xml.org/sax/features/external-general-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/external-parameter-entities"
		    .equals (featureId)
		|| "http://xml.org/sax/features/use-locator"
		    .equals (featureId)) {
	    if (state != true)
		throw new SAXNotSupportedException (featureId);
	    else
		return;
	}
	if ("http://xml.org/sax/features/namespaces"
		    .equals (featureId)
		|| "http://xml.org/sax/features/normalize-text"
		    .equals (featureId)) {
	    if (state != false)
		throw new SAXNotSupportedException (featureId);
	    else
		return;
	}

	// unknown features
	throw new SAXNotRecognizedException (featureId);
    }

    /**
     * <b>SAX2</b>:  Assigns the specified property.  At this time only
     * declaration and lexical handlers are supported, and these must not
     * be changed to values of the wrong type.  Like SAX1 handlers, these
     * may be changed at any time.
     *
     * @exception SAXException as defined in the specification for
     *	<em>org.xml.sax.Configurable.setProperty()</em>
     */
    public void setProperty (String propertyId, Object property)
    throws SAXException
    {
	if ("http://xml.org/sax/handlers/DeclHandler".equals (propertyId)) {
	    if (!(property instanceof DeclHandler))
		throw new SAXNotSupportedException (propertyId);
	    declHandler = (DeclHandler) property;
	    return;
	}
	if ("http://xml.org/sax/handlers/LexicalHandler".equals (propertyId)) {
	    if (!(property instanceof LexicalHandler))
		throw new SAXNotSupportedException (propertyId);
	    lexicalHandler = (LexicalHandler) property;
	    return;
	}

	// unknown properties
	throw new SAXNotRecognizedException (propertyId);
    }

    //
    // This inner class is a little adapter between the events provided by
    // Sun's parsers (based on extremely early proposals for what SAX2
    // should be!) and the alpha SAX2 ones.
    //
    final class Adapter implements
	DocumentHandler, LexicalEventListener,
	DTDHandler, DtdEventListener
    {
	// only set within DTD
	private String		rootName;
	private boolean		startedDtd;
	private boolean		startedExternalSubset;

	// temp used in attribute decls
	private StringBuffer	temp = new StringBuffer ();

	// used to detect otherwise unnamed "[dtd]" PE
	private String		extSubsetSysId;
	private Locator		locator;

	Adapter () { }

	//
	// SAX 1.0 DocumentHandler callbacks
	//	... no loss of information here
	//
	public void setDocumentLocator (Locator locator)
	{
	    this.locator = locator;
	    docHandler.setDocumentLocator (locator);
	}
    
	public void startDocument ()
	throws SAXException
	    { docHandler.startDocument (); }
    
	public void endDocument ()
	throws SAXException
	    { docHandler.endDocument (); }
    
	public void startElement (String name, AttributeList attrs)
	throws SAXException
	    { docHandler.startElement (name, attrs); }
    
	public void endElement (String name)
	throws SAXException
	    { docHandler.endElement (name); }
    
	public void characters (char buf [], int off, int len)
	throws SAXException
	    { docHandler.characters (buf, off, len); }

	public void ignorableWhitespace (char buf [], int off, int len)
	throws SAXException
	    { docHandler.ignorableWhitespace (buf, off, len); }

	public void processingInstruction (String target, String data)
	throws SAXException
	{
	    checkExternalSubset ();
	    docHandler.processingInstruction (target, data);
	}
	
	//
	// Sun LexicalEventListener callbacks
	//	... Sun's parsers don't report PEs at all, but they're
	//	    visible through the locator in callbacks they cause.
	//
	public void startCDATA ()
	throws SAXException
	    { lexicalHandler.startCDATA (); }

	public void endCDATA ()
	throws SAXException
	    { lexicalHandler.endCDATA (); }

	public void comment (String data)
	throws SAXException
	{
	    checkExternalSubset ();
	    lexicalHandler.comment (data.toCharArray (), 0, data.length ());
	}

	public void startParsedEntity (String name)
	throws SAXException
	{
	    // Sun's parser does NOT report parameter entities, since
	    // there's virtually nothing useful to be done with them in
	    // common cases like using them to construct declarations from
	    // fragments, or triggering handling of conditional sections.
	    lexicalHandler.startEntity (name);
	}

	public void endParsedEntity (String name, boolean ignored)
	throws SAXException
	{
	    lexicalHandler.endEntity (name);
	}

	//
	// SAX 1.0 DTDHandler callbacks:
	//	... no loss of information here
	//
	private void checkExternalSubset ()
	throws SAXException
	{
	    if (startedExternalSubset)
		return;
	    if (extSubsetSysId == null
		    || !extSubsetSysId.equals (locator.getSystemId ()))
		return;
	    lexicalHandler.startEntity ("[dtd]");
	    startedExternalSubset = true;
	}

	public void unparsedEntityDecl (String name,
	    String publicId, String systemId,
	    String notation)
	throws SAXException
	{
	    checkExternalSubset ();
	    dtdHandler.unparsedEntityDecl (name, publicId,
		    systemId, notation);
	}

	public void notationDecl (String name,
	    String publicId, String systemId)
	throws SAXException
	{
	    checkExternalSubset ();
	    dtdHandler.notationDecl (name, publicId, systemId);
	}

	//
	// Sun DtdEventListener callbacks
	//	... SAX2 draft discards internal DTD subset
	//	... SAX2 draft can't handle NOTATIONS attributes 
	//
	public void startDtd (String name)
	throws SAXException
	{
	    rootName = name; /* save and use later */
	    extSubsetSysId = null;
	    startedDtd = false;
	}

	public void internalDtdDecl (String subset)
	throws SAXException
	{
	    // SAX2 discards the internal DTD subset (sigh); no real work.
	    if (startedDtd)
		return;

	    // SAX2 groups a mandatory (rootName) part of the DOCTYPE with
	    // an optional (external subset) one, which may never arrive.
	    // This handles the "no external subset" case.
	    lexicalHandler.startDTD (rootName, null, null);
	    startedDtd = true;
	}

	public void externalDtdDecl (String publicId, String systemId)
	throws SAXException
	{
	    lexicalHandler.startDTD (rootName, publicId, systemId);
	    extSubsetSysId = systemId;
	    startedDtd = true;
	}

	public void endDtd ()
	throws SAXException
	{
	    // SAX2 groups a mandatory (rootName) part of the DOCTYPE with
	    // an optional (external subset) one, which may never arrive.
	    // This handles the "neither external nor internal subset" case.
	    if (!startedDtd)
		lexicalHandler.startDTD (rootName, null, null);

	    // SAX2 shows parameter entities, while Sun's parser hides most
	    // of them since seeing their expansions is in general useless.
	    else if (startedExternalSubset)
		lexicalHandler.endEntity ("[dtd]");

	    lexicalHandler.endDTD ();
	    rootName = null;
	}

	public void attributeDecl (
	    String elementName,
	    String attributeName,
	    String type,
	    String options [],
	    String defaultValue,
	    boolean isFixed, boolean isRequired)
	throws SAXException
	{
	    String	valueDefaultType = null;

	    checkExternalSubset ();

	    // avoid the work below, if we clearly can do so
	    if (declHandler == defaultHandler)
		return;

	    if (isRequired)
		valueDefaultType = "#REQUIRED";
	    else if (isFixed)
		valueDefaultType = "#FIXED";
	    else if (defaultValue == null)
		valueDefaultType = "#IMPLIED";
	    
	    //
	    // if "type" is ENUMERATION, SAX2 needs us to
	    // rewrite it into "(option1|option2|...)"
	    //
	    if ("ENUMERATION".equals (type)) {
		temp.append ('(');
		for (int i = 0; i < options.length; i++) {
		    if (i != 0)
			temp.append ('|');
		    temp.append (options [i]);
		}
		temp.append (')');
		type = temp.toString ();
	    }

	    // if "type" is NOTATION, SAX2 discards important information,
	    // namely the particular notations which are permitted!

	    declHandler.attributeDecl (elementName, attributeName,
		type, valueDefaultType, defaultValue);
	}

	public void elementDecl (String name, String model)
	throws SAXException
	{
	    checkExternalSubset ();
	    declHandler.elementDecl (name, model);
	}

	public void externalEntityDecl (String name,
	    String publicId, String systemId)
	throws SAXException
	{
	    // Only general entities, not parameter entities
	    checkExternalSubset ();
	    declHandler.externalEntityDecl (name, publicId, systemId);
	}

	public void internalEntityDecl (String name, String value)
	throws SAXException
	{
	    // Only general entities, not parameter entities
	    checkExternalSubset ();
	    declHandler.internalEntityDecl (name, value);
	}
    }
}
