/*
 * Copyright (c) 1999 by David Brownell.  All Rights Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package org.brownell.xml;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;


/**
 * This extends the EchoHandler to create a class which defaults to writing
 * XHTML text, preferring the US-ASCII encoding.  It adds no unique
 * functionality, only changing the defaults slightly to simplify writing
 * XHTML processing components by providing a bean class whose properties
 * have more convenient defaults.  An artifact of using the US-ASCII
 * encoding is that no XML declaration is written, so that HTML tools
 * that can't accept them will not become confused.  Components can treat
 * the output as UTF-8, ISO-8859-1, or US-ASCII without incurring any
 * data loss.
 *
 * @author David Brownell (db@post.harvard.edu)
 * @version 1.0 (8 September 1999)
 */
public class XhtmlEchoHandler extends EchoHandler
{
    /**
     * Constructs this handler such System.out is used to write
     * SAX events using the US-ASCII encoding, as XHTML.
     */
    public XhtmlEchoHandler ()
    throws IOException
    {
	this (System.out);
    }

    /**
     * Constructs this handler such that the specified output stream
     * is used to write SAX events in the US-ASCII encoding, as XHTML.
     *
     * @param out Where US-ASCII encoding of the stream of SAX
     *	events will be sent.
     */
    public XhtmlEchoHandler (OutputStream out)
    throws IOException
    {
	this (new OutputStreamWriter (out, "ASCII"), "US-ASCII");
    }

    /**
     * Constructs this handler such that the specified output stream
     * is used to write SAX events as XHTML.
     *
     * @param out Where the stream of SAX events will be written.
     */
    public XhtmlEchoHandler (Writer out)
    {
	this (out, null);
    }

    /**
     * Constructs this handler such that the specified output stream
     * is used to write SAX events as XHTML, labeled with the specified
     * encoding.
     *
     * @param out Where the stream of SAX events will be written.
     * @param encoding If non-null, this names the encoding to be
     *	placed in the encoding declaration.
     */
    public XhtmlEchoHandler (Writer out, String encoding)
    {
	super (out, encoding);
	setXhtml (true);
    }
}
